import re
from typing import Iterable, Sequence, Union

from pyglet import gl
from typing_extensions import TypeAlias

from arcade.types import BufferProtocol

from .buffer import Buffer

BufferOrBufferProtocol = Union[BufferProtocol, Buffer]

GLenumLike = Union[gl.GLenum, int]
PyGLenum = int
GLuintLike = Union[gl.GLuint, int]
PyGLuint = int


OpenGlFilter: TypeAlias = tuple[PyGLenum, PyGLenum]
BlendFunction: TypeAlias = Union[
    tuple[PyGLenum, PyGLenum], tuple[PyGLenum, PyGLenum, PyGLenum, PyGLenum]
]

#: Depth compare functions
compare_funcs: dict[str | None, int] = {
    None: gl.GL_NONE,
    "<=": gl.GL_LEQUAL,
    "<": gl.GL_LESS,
    ">=": gl.GL_GEQUAL,
    ">": gl.GL_GREATER,
    "==": gl.GL_EQUAL,
    "!=": gl.GL_NOTEQUAL,
    "0": gl.GL_NEVER,
    "1": gl.GL_ALWAYS,
}

#: Swizzle conversion lookup
swizzle_enum_to_str: dict[int, str] = {
    gl.GL_RED: "R",
    gl.GL_GREEN: "G",
    gl.GL_BLUE: "B",
    gl.GL_ALPHA: "A",
    gl.GL_ZERO: "0",
    gl.GL_ONE: "1",
}

#: Swizzle conversion lookup
swizzle_str_to_enum: dict[str, int] = {
    "R": gl.GL_RED,
    "G": gl.GL_GREEN,
    "B": gl.GL_BLUE,
    "A": gl.GL_ALPHA,
    "0": gl.GL_ZERO,
    "1": gl.GL_ONE,
}

_float_base_format = (0, gl.GL_RED, gl.GL_RG, gl.GL_RGB, gl.GL_RGBA)
_int_base_format = (
    0,
    gl.GL_RED_INTEGER,
    gl.GL_RG_INTEGER,
    gl.GL_RGB_INTEGER,
    gl.GL_RGBA_INTEGER,
)
#: Pixel format lookup (base_format, internal_format, type, size)
pixel_formats = {
    # float formats
    "f1": (
        _float_base_format,
        (0, gl.GL_R8, gl.GL_RG8, gl.GL_RGB8, gl.GL_RGBA8),
        gl.GL_UNSIGNED_BYTE,
        1,
    ),
    "f2": (
        _float_base_format,
        (0, gl.GL_R16F, gl.GL_RG16F, gl.GL_RGB16F, gl.GL_RGBA16F),
        gl.GL_HALF_FLOAT,
        2,
    ),
    "f4": (
        _float_base_format,
        (0, gl.GL_R32F, gl.GL_RG32F, gl.GL_RGB32F, gl.GL_RGBA32F),
        gl.GL_FLOAT,
        4,
    ),
    # int formats
    "i1": (
        _int_base_format,
        (0, gl.GL_R8I, gl.GL_RG8I, gl.GL_RGB8I, gl.GL_RGBA8I),
        gl.GL_BYTE,
        1,
    ),
    "i2": (
        _int_base_format,
        (0, gl.GL_R16I, gl.GL_RG16I, gl.GL_RGB16I, gl.GL_RGBA16I),
        gl.GL_SHORT,
        2,
    ),
    "i4": (
        _int_base_format,
        (0, gl.GL_R32I, gl.GL_RG32I, gl.GL_RGB32I, gl.GL_RGBA32I),
        gl.GL_INT,
        4,
    ),
    # uint formats
    "u1": (
        _int_base_format,
        (0, gl.GL_R8UI, gl.GL_RG8UI, gl.GL_RGB8UI, gl.GL_RGBA8UI),
        gl.GL_UNSIGNED_BYTE,
        1,
    ),
    "u2": (
        _int_base_format,
        (0, gl.GL_R16UI, gl.GL_RG16UI, gl.GL_RGB16UI, gl.GL_RGBA16UI),
        gl.GL_UNSIGNED_SHORT,
        2,
    ),
    "u4": (
        _int_base_format,
        (0, gl.GL_R32UI, gl.GL_RG32UI, gl.GL_RGB32UI, gl.GL_RGBA32UI),
        gl.GL_UNSIGNED_INT,
        4,
    ),
}


#: String representation of a shader types
SHADER_TYPE_NAMES = {
    gl.GL_VERTEX_SHADER: "vertex shader",
    gl.GL_FRAGMENT_SHADER: "fragment shader",
    gl.GL_GEOMETRY_SHADER: "geometry shader",
    gl.GL_TESS_CONTROL_SHADER: "tessellation control shader",
    gl.GL_TESS_EVALUATION_SHADER: "tessellation evaluation shader",
}

#: Lookup table for OpenGL type names
GL_NAMES = {
    gl.GL_HALF_FLOAT: "GL_HALF_FLOAT",
    gl.GL_FLOAT: "GL_FLOAT",
    gl.GL_DOUBLE: "GL_DOUBLE",
    gl.GL_INT: "GL_INT",
    gl.GL_UNSIGNED_INT: "GL_UNSIGNED_INT",
    gl.GL_SHORT: "GL_SHORT",
    gl.GL_UNSIGNED_SHORT: "GL_UNSIGNED_SHORT",
    gl.GL_BYTE: "GL_BYTE",
    gl.GL_UNSIGNED_BYTE: "GL_UNSIGNED_BYTE",
}


def gl_name(gl_type: PyGLenum | None) -> str | PyGLenum | None:
    """Return the name of a gl type"""
    if gl_type is None:
        return None
    return GL_NAMES.get(gl_type, gl_type)


class AttribFormat:
    """
    Represents a vertex attribute in a BufferDescription / Program.
    This is attribute metadata used when attempting to map vertex
    shader inputs.

    Args:
        name:
            Name of the attribute
        gl_type:
            The OpenGL type such as GL_FLOAT, GL_HALF_FLOAT etc.
        bytes_per_component:
            Number of bytes for a single component
        offset:
            Offset for BufferDescription
        location:
            Location for program attribute
    """

    __slots__ = (
        "name",
        "gl_type",
        "components",
        "bytes_per_component",
        "offset",
        "location",
    )

    def __init__(
        self,
        name: str | None,
        gl_type: PyGLenum | None,
        components: int,
        bytes_per_component: int,
        offset=0,
        location=0,
    ):
        self.name = name
        """The name of the attribute in the program"""
        self.gl_type = gl_type
        """The OpenGL type of the attribute"""
        self.components = components
        """Number of components for this attribute (1, 2, 3 or 4)"""
        self.bytes_per_component = bytes_per_component
        """How many bytes for a single component"""
        self.offset = offset
        """Offset of the attribute in the buffer"""
        self.location = location
        """Location of the attribute in the program"""

    @property
    def bytes_total(self) -> int:
        """Total number of bytes for this attribute"""
        return self.components * self.bytes_per_component

    def __repr__(self):
        return (
            f"<AttribFormat {self.name} {self.gl_type} components={self.components} "
            f"bytes_per_component={self.bytes_per_component}>"
        )


class BufferDescription:
    """Buffer Object description used with :py:class:`arcade.gl.Geometry`.

    This class provides a Buffer object with a description of its content, allowing the
    a :py:class:`~arcade.gl.Geometry` object to correctly map shader attributes
    to a program/shader.

    The formats is a string providing the number and type of each attribute. Currently
    we only support f (float), i (integer) and B (unsigned byte).

    ``normalized`` enumerates the attributes which must have their values normalized.
    This is useful for instance for colors attributes given as unsigned byte and
    normalized to floats with values between 0.0 and 1.0.

    ``instanced`` allows this buffer to be used as instanced buffer. Each value will
    be used once for the whole geometry. The geometry will be repeated a number of
    times equal to the number of items in the Buffer.

    .. code-block:: python

        # Describe my_buffer
        # It contains two floating point numbers being a 2d position
        # and two floating point numbers being texture coordinates.
        # We expect the shader using this buffer to have an in_pos and in_uv attribute (exact name)
        BufferDescription(
            my_buffer,
            '2f 2f',
            ['in_pos', 'in_uv'],
        )

    Args:
        buffer: The buffer to describe
        formats: The format of each attribute
        attributes: List of attributes names (strings)
        normalized: list of attribute names that should be normalized
        instanced: ``True`` if this is per instance data
    """

    # Describe all variants of a format string to simplify parsing (single component)
    # format: gl_type, byte_size
    _formats: dict[str, tuple[PyGLenum | None, int]] = {
        # (gl enum, byte size)
        # Floats
        "f": (gl.GL_FLOAT, 4),
        "f1": (gl.GL_UNSIGNED_BYTE, 1),
        "f2": (gl.GL_HALF_FLOAT, 2),
        "f4": (gl.GL_FLOAT, 4),
        "f8": (gl.GL_DOUBLE, 8),
        # Unsigned integers
        "u": (gl.GL_UNSIGNED_INT, 4),
        "u1": (gl.GL_UNSIGNED_BYTE, 1),
        "u2": (gl.GL_UNSIGNED_SHORT, 2),
        "u4": (gl.GL_UNSIGNED_INT, 4),
        # Signed integers
        "i": (gl.GL_INT, 4),
        "i1": (gl.GL_BYTE, 1),
        "i2": (gl.GL_SHORT, 2),
        "i4": (gl.GL_INT, 4),
        # Padding (1, 2, 4, 8 bytes)
        "x": (None, 1),
        "x1": (None, 1),
        "x2": (None, 2),
        "x4": (None, 4),
        "x8": (None, 8),
    }

    __slots__ = (
        "buffer",
        "attributes",
        "normalized",
        "instanced",
        "formats",
        "stride",
        "num_vertices",
    )

    def __init__(
        self,
        buffer: Buffer,
        formats: str,
        attributes: Sequence[str],
        normalized: Iterable[str] | None = None,
        instanced: bool = False,
    ):
        #: The :py:class:`~arcade.gl.Buffer` this description object describes
        self.buffer = buffer  # type: Buffer
        #: List of string attributes
        self.attributes = attributes
        #: List of normalized attributes
        self.normalized: set[str] = set() if normalized is None else set(normalized)
        #: Instanced flag (bool)
        self.instanced: bool = instanced
        #: Formats of each attribute
        self.formats: list[AttribFormat] = []
        #: The byte stride of the buffer
        self.stride: int = -1
        #: Number of vertices in the buffer
        self.num_vertices: int = -1

        if not isinstance(buffer, Buffer):
            raise ValueError("buffer parameter must be an arcade.gl.Buffer")

        if not isinstance(self.attributes, (list, tuple)):
            raise ValueError("Attributes must be a list or tuple")

        if self.normalized > set(self.attributes):
            raise ValueError("Normalized attribute not found in attributes.")

        formats_list = formats.split(" ")
        non_padded_formats = [f for f in formats_list if "x" not in f]

        if len(non_padded_formats) != len(self.attributes):
            raise ValueError(
                f"Different lengths of formats ({len(non_padded_formats)}) and "
                f"attributes ({len(self.attributes)})"
            )

        def zip_attrs(formats: list[str], attributes: Sequence[str]):
            """Join together formats and attribute names taking padding into account"""
            attr_index = 0
            for f in formats:
                if "x" in f:
                    yield f, None
                else:
                    yield f, attributes[attr_index]
                    attr_index += 1

        self.stride = 0
        for attr_fmt, attr_name in zip_attrs(formats_list, self.attributes):
            # Automatically make f1 attributes normalized
            if attr_name is not None and "f1" in attr_fmt:
                self.normalized.add(attr_name)
            try:
                components_str, data_type_str, data_size_str = re.split(r"([fiux])", attr_fmt)
                data_type = f"{data_type_str}{data_size_str}" if data_size_str else data_type_str
                components = int(components_str) if components_str else 1  # 1 component is default
                data_size = (
                    int(data_size_str) if data_size_str else 4
                )  # 4 byte float and integer types are default
                # Limit components to 4 for non-padded formats
                if components > 4 and data_size is not None:
                    raise ValueError("Number of components must be 1, 2, 3 or 4")
            except Exception as ex:
                raise ValueError(f"Could not parse attribute format: '{attr_fmt} : {ex}'")

            gl_type, byte_size = self._formats[data_type]
            self.formats.append(
                AttribFormat(attr_name, gl_type, components, byte_size, offset=self.stride)
            )

            self.stride += byte_size * components

        if self.buffer.size % self.stride != 0:
            raise ValueError(
                f"Buffer size must align by {self.stride} bytes. "
                f"{self.buffer} size={self.buffer.size}"
            )

        # Estimate number of vertices for this buffer
        self.num_vertices = self.buffer.size // self.stride

    def __repr__(self) -> str:
        return f"<BufferDescription {self.attributes} {self.formats}>"

    def __eq__(self, other) -> bool:
        if not isinstance(other, BufferDescription):
            raise ValueError(
                f"The only logical comparison to a BufferDescription"
                f"is a BufferDescription not {type(other)}"
            )

        # Equal if we share the same attribute
        return len(set(self.attributes) & set(other.attributes)) > 0


class TypeInfo:
    """
    Describes an opengl type.

    Args:
        name:
            The string representation of this type
        enum:
            The enum of this type
        gl_type:
            The base enum of this type
        gl_size:
            byte size if the gl_type
        components:
            Number of components for this enum
    """

    __slots__ = "name", "enum", "gl_type", "gl_size", "components"

    def __init__(
        self, name: str, enum: GLenumLike, gl_type: PyGLenum, gl_size: int, components: int
    ):
        self.name = name
        """The string representation of this type"""
        self.enum = enum
        """The OpenEL enum of this type"""
        self.gl_type = gl_type
        """The base OpenGL data type"""
        self.gl_size = gl_size
        """The size of the base OpenGL data type"""
        self.components = components
        """The number of components (1, 2, 3 or 4)"""

    @property
    def size(self) -> int:
        """The total size of this type in bytes"""
        return self.gl_size * self.components

    def __repr__(self) -> str:
        return (
            f"<TypeInfo name={self.name}, enum={self.enum} gl_type={self.gl_type} "
            f"gl_size={self.gl_size} components={self.components}>"
        )


class GLTypes:
    """
    Detailed Information about all attribute type.

    During introspection we often just get integers telling us what type is used.
    This can for example be ``35664`` telling us it's a ``GL_FLOAT_VEC2``.

    During introspection we need to know the exact datatype of the attribute.
    It's not enough to know it's a float, we need to know if it's a vec2, vec3, vec4
    or any other type that OpenGL supports.

    Examples of types are::

        GL_FLOAT_VEC2
        GL_DOUBLE_VEC4
        GL_INT_VEC3
        GL_UNSIGNED_INT_VEC2
        GL_UNSIGNED_BYTE
        GL_FLOAT
        GL_DOUBLE
        GL_INT
        GL_UNSIGNED_INT
        ...
    """

    types = {
        # Floats
        gl.GL_FLOAT: TypeInfo("GL_FLOAT", gl.GL_FLOAT, gl.GL_FLOAT, 4, 1),
        gl.GL_FLOAT_VEC2: TypeInfo("GL_FLOAT_VEC2", gl.GL_FLOAT_VEC2, gl.GL_FLOAT, 4, 2),
        gl.GL_FLOAT_VEC3: TypeInfo("GL_FLOAT_VEC3", gl.GL_FLOAT_VEC3, gl.GL_FLOAT, 4, 3),
        gl.GL_FLOAT_VEC4: TypeInfo("GL_FLOAT_VEC4", gl.GL_FLOAT_VEC4, gl.GL_FLOAT, 4, 4),
        # Doubles
        gl.GL_DOUBLE: TypeInfo("GL_DOUBLE", gl.GL_DOUBLE, gl.GL_DOUBLE, 8, 1),
        gl.GL_DOUBLE_VEC2: TypeInfo("GL_DOUBLE_VEC2", gl.GL_DOUBLE_VEC2, gl.GL_DOUBLE, 8, 2),
        gl.GL_DOUBLE_VEC3: TypeInfo("GL_DOUBLE_VEC3", gl.GL_DOUBLE_VEC3, gl.GL_DOUBLE, 8, 3),
        gl.GL_DOUBLE_VEC4: TypeInfo("GL_DOUBLE_VEC4", gl.GL_DOUBLE_VEC4, gl.GL_DOUBLE, 8, 4),
        # Booleans (ubyte)
        gl.GL_BOOL: TypeInfo("GL_BOOL", gl.GL_BOOL, gl.GL_BOOL, 1, 1),
        gl.GL_BOOL_VEC2: TypeInfo("GL_BOOL_VEC2", gl.GL_BOOL_VEC2, gl.GL_BOOL, 1, 2),
        gl.GL_BOOL_VEC3: TypeInfo("GL_BOOL_VEC3", gl.GL_BOOL_VEC3, gl.GL_BOOL, 1, 3),
        gl.GL_BOOL_VEC4: TypeInfo("GL_BOOL_VEC4", gl.GL_BOOL_VEC4, gl.GL_BOOL, 1, 4),
        # Integers
        gl.GL_INT: TypeInfo("GL_INT", gl.GL_INT, gl.GL_INT, 4, 1),
        gl.GL_INT_VEC2: TypeInfo("GL_INT_VEC2", gl.GL_INT_VEC2, gl.GL_INT, 4, 2),
        gl.GL_INT_VEC3: TypeInfo("GL_INT_VEC3", gl.GL_INT_VEC3, gl.GL_INT, 4, 3),
        gl.GL_INT_VEC4: TypeInfo("GL_INT_VEC4", gl.GL_INT_VEC4, gl.GL_INT, 4, 4),
        # Unsigned Integers
        gl.GL_UNSIGNED_INT: TypeInfo(
            "GL_UNSIGNED_INT", gl.GL_UNSIGNED_INT, gl.GL_UNSIGNED_INT, 4, 1
        ),
        gl.GL_UNSIGNED_INT_VEC2: TypeInfo(
            "GL_UNSIGNED_INT_VEC2", gl.GL_UNSIGNED_INT_VEC2, gl.GL_UNSIGNED_INT, 4, 2
        ),
        gl.GL_UNSIGNED_INT_VEC3: TypeInfo(
            "GL_UNSIGNED_INT_VEC3", gl.GL_UNSIGNED_INT_VEC3, gl.GL_UNSIGNED_INT, 4, 3
        ),
        gl.GL_UNSIGNED_INT_VEC4: TypeInfo(
            "GL_UNSIGNED_INT_VEC4", gl.GL_UNSIGNED_INT_VEC4, gl.GL_UNSIGNED_INT, 4, 4
        ),
        # Unsigned Short (mostly used for short index buffers)
        gl.GL_UNSIGNED_SHORT: TypeInfo(
            "GL.GL_UNSIGNED_SHORT", gl.GL_UNSIGNED_SHORT, gl.GL_UNSIGNED_SHORT, 2, 2
        ),
        # Byte
        gl.GL_BYTE: TypeInfo("GL_BYTE", gl.GL_BYTE, gl.GL_BYTE, 1, 1),
        gl.GL_UNSIGNED_BYTE: TypeInfo(
            "GL_UNSIGNED_BYTE", gl.GL_UNSIGNED_BYTE, gl.GL_UNSIGNED_BYTE, 1, 1
        ),
        # Matrices
        gl.GL_FLOAT_MAT2: TypeInfo("GL_FLOAT_MAT2", gl.GL_FLOAT_MAT2, gl.GL_FLOAT, 4, 4),
        gl.GL_FLOAT_MAT3: TypeInfo("GL_FLOAT_MAT3", gl.GL_FLOAT_MAT3, gl.GL_FLOAT, 4, 9),
        gl.GL_FLOAT_MAT4: TypeInfo("GL_FLOAT_MAT4", gl.GL_FLOAT_MAT4, gl.GL_FLOAT, 4, 16),
        gl.GL_FLOAT_MAT2x3: TypeInfo("GL_FLOAT_MAT2x3", gl.GL_FLOAT_MAT2x3, gl.GL_FLOAT, 4, 6),
        gl.GL_FLOAT_MAT2x4: TypeInfo("GL_FLOAT_MAT2x4", gl.GL_FLOAT_MAT2x4, gl.GL_FLOAT, 4, 8),
        gl.GL_FLOAT_MAT3x2: TypeInfo("GL_FLOAT_MAT3x2", gl.GL_FLOAT_MAT3x2, gl.GL_FLOAT, 4, 6),
        gl.GL_FLOAT_MAT3x4: TypeInfo("GL_FLOAT_MAT3x4", gl.GL_FLOAT_MAT3x4, gl.GL_FLOAT, 4, 12),
        gl.GL_FLOAT_MAT4x2: TypeInfo("GL_FLOAT_MAT4x2", gl.GL_FLOAT_MAT4x2, gl.GL_FLOAT, 4, 8),
        gl.GL_FLOAT_MAT4x3: TypeInfo("GL_FLOAT_MAT4x3", gl.GL_FLOAT_MAT4x3, gl.GL_FLOAT, 4, 12),
        # Double matrices
        gl.GL_DOUBLE_MAT2: TypeInfo("GL_DOUBLE_MAT2", gl.GL_DOUBLE_MAT2, gl.GL_DOUBLE, 8, 4),
        gl.GL_DOUBLE_MAT3: TypeInfo("GL_DOUBLE_MAT3", gl.GL_DOUBLE_MAT3, gl.GL_DOUBLE, 8, 9),
        gl.GL_DOUBLE_MAT4: TypeInfo("GL_DOUBLE_MAT4", gl.GL_DOUBLE_MAT4, gl.GL_DOUBLE, 8, 16),
        gl.GL_DOUBLE_MAT2x3: TypeInfo("GL_DOUBLE_MAT2x3", gl.GL_DOUBLE_MAT2x3, gl.GL_DOUBLE, 8, 6),
        gl.GL_DOUBLE_MAT2x4: TypeInfo("GL_DOUBLE_MAT2x4", gl.GL_DOUBLE_MAT2x4, gl.GL_DOUBLE, 8, 8),
        gl.GL_DOUBLE_MAT3x2: TypeInfo("GL_DOUBLE_MAT3x2", gl.GL_DOUBLE_MAT3x2, gl.GL_DOUBLE, 8, 6),
        gl.GL_DOUBLE_MAT3x4: TypeInfo("GL_DOUBLE_MAT3x4", gl.GL_DOUBLE_MAT3x4, gl.GL_DOUBLE, 8, 12),
        gl.GL_DOUBLE_MAT4x2: TypeInfo("GL_DOUBLE_MAT4x2", gl.GL_DOUBLE_MAT4x2, gl.GL_DOUBLE, 8, 8),
        gl.GL_DOUBLE_MAT4x3: TypeInfo("GL_DOUBLE_MAT4x3", gl.GL_DOUBLE_MAT4x3, gl.GL_DOUBLE, 8, 12),
        # TODO: Add sampler types if needed. Only needed for better uniform introspection.
    }

    @classmethod
    def get(cls, enum: int) -> TypeInfo:
        """Get the TypeInfo for a given"""
        try:
            return cls.types[enum]
        except KeyError:
            raise ValueError(f"Unknown gl type {enum}. Someone needs to add it")



from array import array
from collections.abc import Iterator, Sequence

from arcade import gl
from arcade.color import WHITE
from arcade.future.texture_render_target import RenderTargetTexture
from arcade.types import RGBOrA255


class Light:
    """
    Create a Light.

    Note: It's important to separate lights that don't change properties
    and static ones with the ``usage`` parameter.

    Args:
        center_x: X position of the light
        center_y: Y position of the light
        radius: Radius of the light
        color: Color of the light
        mode: 'hard' or 'soft' light
    """

    HARD = 1.0
    SOFT = 0.0

    def __init__(
        self,
        center_x: float,
        center_y: float,
        radius: float = 50.0,
        color: RGBOrA255 = WHITE,
        mode: str = "hard",
    ):
        if not (isinstance(color, tuple) or isinstance(color, list)):
            raise ValueError(
                "Color must be a 3-4 element Tuple or List with red-green-blue "
                " and optionally an alpha."
            )

        if not isinstance(mode, str) or not (mode == "soft" or mode == "hard"):
            raise ValueError("Mode must be set to either 'soft' or 'hard'.")

        self._center_x = center_x
        self._center_y = center_y
        self._radius = radius
        self._attenuation = Light.HARD if mode == "hard" else Light.SOFT
        self._color = color[:3]
        self._light_layer: LightLayer | None = None

        if len(self._color) != 3:
            raise ValueError(
                "Color must be a 3-4 element Tuple or List with red-green-blue "
                "and optionally an alpha."
            )

    @property
    def position(self) -> tuple[float, float]:
        """Get or set the light position"""
        return self._center_x, self._center_y

    @position.setter
    def position(self, value):
        if self._light_layer:
            self._light_layer._rebuild = True
        self._center_x, self._center_y = value

    @property
    def radius(self) -> float:
        """Get or set the light size"""
        return self._radius

    @radius.setter
    def radius(self, value):
        if self._light_layer:
            self._light_layer._rebuild = True
        self._radius = value


class LightLayer(RenderTargetTexture):
    """
    Create a LightLayer

    The size of a layer should ideally be of the same size and the screen.

    Args:
        width: Width of light layer
        height: Height of light layer
    """

    def __init__(self, width: int, height: int):
        super().__init__(width, height)

        self._lights: list[Light] = []
        self._prev_target = None
        self._rebuild = False
        self._stride = 28
        self._buffer = self.ctx.buffer(reserve=self._stride * 100)
        self._vao = self.ctx.geometry(
            [
                gl.BufferDescription(
                    self._buffer,
                    "2f 1f 1f 3f",
                    ["in_vert", "in_radius", "in_attenuation", "in_color"],
                ),
            ]
        )
        self._light_program = self.ctx.load_program(
            vertex_shader=":system:shaders/lights/point_lights_vs.glsl",
            geometry_shader=":system:shaders/lights/point_lights_geo.glsl",
            fragment_shader=":system:shaders/lights/point_lights_fs.glsl",
        )
        self._combine_program = self.ctx.load_program(
            vertex_shader=":system:shaders/lights/combine_vs.glsl",
            fragment_shader=":system:shaders/lights/combine_fs.glsl",
        )
        # NOTE: Diffuse buffer created in parent
        self._light_buffer = self.ctx.framebuffer(
            color_attachments=self.ctx.texture((width, height), components=3)
        )

    @property
    def diffuse_texture(self):
        """The diffuse texture"""
        return self.texture

    @property
    def light_texture(self):
        """The light texture"""
        return self._light_buffer.color_attachments[0]

    def resize(self, width, height):
        """Resize the light layer"""
        super().resize(width, height)
        self._light_buffer = self.ctx.framebuffer(
            color_attachments=self.ctx.texture((width, height), components=3)
        )

    def clear(self):
        """Clear the light layer"""
        super().clear()
        self._light_buffer.clear()

    def add(self, light: Light):
        """Add a Light to the layer"""
        self._lights.append(light)
        light._light_layer = self
        self._rebuild = True

    def extend(self, lights: Sequence[Light]):
        """Add a list of lights to the layer"""
        for light in lights:
            self.add(light)

    def remove(self, light: Light):
        """Remove a light to the layer"""
        self._lights.remove(light)
        light._light_layer = None
        self._rebuild = True

    def __len__(self) -> int:
        """Number of lights"""
        return len(self._lights)

    def __iter__(self) -> Iterator[Light]:
        """Return an iterable object of lights"""
        return iter(self._lights)

    def __getitem__(self, i) -> Light:
        return self._lights[i]

    def __enter__(self):
        self._prev_target = self.ctx.active_framebuffer
        self._fbo.use()
        self._fbo.clear(color=self._background_color)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._prev_target.use()

    def draw(
        self,
        position: tuple[float, float] = (0, 0),
        target=None,
        ambient_color: RGBOrA255 = (64, 64, 64),
    ):
        """Draw the lights

        Args:
            position: Position offset (scrolling)
            target: The window or framebuffer we want to render to (default is window)
            ambient_color: The ambient light color
        """
        if target is None:
            target = self.window

        # Re-build light data if needed
        if self._rebuild and len(self._lights) > 0:
            data: list[float] = []
            for light in self._lights:
                data.extend(light.position)
                data.append(light.radius)
                data.append(light._attenuation)
                data.extend(light._color)

            while self._buffer.size < len(data) * self._stride:
                self._buffer.orphan(double=True)

            self._buffer.write(data=array("f", data))
            self._rebuild = False

        # Render to light buffer
        self._light_buffer.use()
        self._light_buffer.clear()
        if len(self._lights) > 0:
            self._light_program["position"] = position
            self.ctx.enable(self.ctx.BLEND)
            self.ctx.blend_func = self.ctx.BLEND_ADDITIVE
            self._vao.render(self._light_program, mode=self.ctx.POINTS, vertices=len(self._lights))
            self.ctx.blend_func = self.ctx.BLEND_DEFAULT

        # Combine pass
        target.use()
        self._combine_program["diffuse_buffer"] = 0
        self._combine_program["light_buffer"] = 1
        self._combine_program["ambient"] = ambient_color[:3]
        self._fbo.color_attachments[0].use(0)
        self._light_buffer.color_attachments[0].use(1)

        self._quad_fs.render(self._combine_program)



"""
Custom Arcade errors, exceptions, and warnings.
"""

import functools
import warnings
from typing import TypeVar

__all__ = [
    "OutsideRangeError",
    "IntOutsideRangeError",
    "FloatOutsideRangeError",
    "ByteRangeError",
    "NormalizedRangeError",
    "PerformanceWarning",
    "ReplacementWarning",
    "warning",
]

# Since this module forbids importing from the rest of
# Arcade, we make our own local type variables.
_TType = TypeVar("_TType", bound=type)
_CT = TypeVar("_CT")  # Comparable type, ie supports the <= operator


class OutsideRangeError(ValueError):
    """
    Raised when a value is outside and expected range

    This class and its subclasses are intended to be Arcade-internal
    helpers to clearly signal exactly what went wrong. Each helps
    type annotate and template a string describing exactly what went
    wrong.

    Args:
        var_name: The name of the variable or argument
        value: The value that fell outside the expected range
        lower: The lower bound, inclusive, of the range
        upper: The upper bound, inclusive, of the range
    """

    def __init__(self, var_name: str, value: _CT, lower: _CT, upper: _CT) -> None:
        super().__init__(f"{var_name} must be between {lower} and {upper}, inclusive, not {value}")
        self.var_name = var_name
        self.value = value
        self.lower = lower
        self.upper = upper


class IntOutsideRangeError(OutsideRangeError):
    """
    An integer was outside an expected range

    This class was originally intended to assist deserialization from
    data packed into ints, such as :py:class:`~arcade.types.Color`.

    Args:
        var_name: The name of the variable or argument
        value: The value that fell outside the expected range
        lower: The lower bound, inclusive, of the range
        upper: The upper bound, inclusive, of the range
    """

    def __init__(self, var_name: str, value: int, lower: int, upper: int) -> None:
        super().__init__(var_name, value, lower, upper)


class FloatOutsideRangeError(OutsideRangeError):
    """
    A float value was outside an expected range

    Args:
        var_name: The name of the variable or argument
        value: The value that fell outside the expected range
        lower: The lower bound, inclusive, of the range
        upper: The upper bound, inclusive, of the range
    """

    def __init__(self, var_name: str, value: float, lower: float, upper: float) -> None:
        super().__init__(var_name, value, lower, upper)


class ByteRangeError(IntOutsideRangeError):
    """
    An int was outside the range of 0 to 255 inclusive

    Args:
        var_name: the name of the variable or argument
        value: the value to fall outside the expected range
    """

    def __init__(self, var_name: str, value: int) -> None:
        super().__init__(var_name, value, 0, 255)


class NormalizedRangeError(FloatOutsideRangeError):
    """
    A float was not between 0.0 and 1.0, inclusive

    Note that normalized floats should not normally be bound-checked as
    before drawing as this is taken care of on the GPU side.

    The exceptions to this are when processing data on the Python side,
    especially when it is cheaper to bound check two floats than call
    clamping functions.

    Args:
        var_name: the name of the variable or argument
        value: the value to fall outside the expected range
    """

    def __init__(self, var_name: str, value: float) -> None:
        super().__init__(var_name, value, 0.0, 1.0)


class PerformanceWarning(Warning):
    """Issuing performance warnings on slow paths."""

    pass


class ReplacementWarning(Warning):
    """Issuing warnings about naming and functionality changes."""

    pass


def warning(warning_type: type[Warning], message: str = "", **kwargs):
    """Warning decorator"""

    def actual_warning_decorator(func):
        nonlocal message
        if warning_type == ReplacementWarning and not message:
            message = f"{func.__name__} is deprecated. Use {kwargs.get('new_name', '')} instead."

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(message, warning_type)
            return func(*args, **kwargs)

        return wrapper

    return actual_warning_decorator



"""Demonstrates the use of layouts.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.1_layouts

Content:
- Create a view with a description of layouts
- Use UIAnchorLayout to position widgets relative to the screen edges
- Use UIBoxLayout to position widgets in a horizontal or vertical line
- Use UIGridLayout to position widgets in a grid

"""

from datetime import datetime

import arcade
from arcade.gui import UIAnchorLayout, UIImage, UITextArea

arcade.resources.load_kenney_fonts()

DESCRIPTION = """How to place widgets on the screen?

Widgets can be manually placed on the screen by using the `add` method of the UIManager
and setting the rect of a widget. This is useful for simple ui setups.

This approach requires to manually calculate the position of the
widgets and handling screen resizes.

Another approach is to use layouts. Layouts are containers that automatically
position widgets based on the layout rules. This example shows how to use
the UIAnchorLayout, UIBoxLayout, and UIGridLayout.

UIAnchorLayout:
- UIAnchorLayout is a layout that positions widgets relative to the screen edges.
- Widgets can be anchored to the top, bottom, left, right, center, or any combination of these.
- Widgets can be aligned to the anchor position with a pixel offset.

UIBoxLayout:
- UIBoxLayout is a layout that positions widgets in a horizontal or vertical line.
- Widgets can be aligned on the orthogonal axis.
- Widgets can have a space between them.

UIGridLayout:
- UIGridLayout is a layout that positions widgets in a grid.
- Widgets can be placed in rows and columns.
- Widgets can have a vertical and horizontal spacing.

The layouts calculate the size of the widgets based on the size_hint,
size_hint_min, and size_hint_max.

- size_hint: A tuple of two floats that represent the relative size of the widget.
- size_hint_min: A tuple of two floats that represent the minimum size of the widget.
- size_hint_max: A tuple of two floats that represent the maximum size of the widget.

Layouts only resize widgets if the size_hint is set for the resprecitve axis.
If the size_hint is not set, the widget will use its current size.

Some widgets calculate their minimum size based on their content like UILabel
and layouts in general.
"""

TEX_SCROLL_DOWN = arcade.load_texture(":resources:gui_basic_assets/scroll/indicator_down.png")
TEX_SCROLL_UP = arcade.load_texture(":resources:gui_basic_assets/scroll/indicator_up.png")


class ScrollableTextArea(UITextArea, UIAnchorLayout):
    """This widget is a text area that can be scrolled, like a UITextLayout, but shows indicator,
    that the text can be scrolled."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        indicator_size = 22
        self._down_indicator = UIImage(
            texture=TEX_SCROLL_DOWN, size_hint=None, width=indicator_size, height=indicator_size
        )
        self._down_indicator.visible = False
        self.add(self._down_indicator, anchor_x="right", anchor_y="bottom", align_x=-3)

        self._up_indicator = UIImage(
            texture=TEX_SCROLL_UP, size_hint=None, width=indicator_size, height=indicator_size
        )
        self._up_indicator.visible = False
        self.add(self._up_indicator, anchor_x="right", anchor_y="top", align_x=-3)

    def on_update(self, dt):
        self._up_indicator.visible = self.layout.view_y < 0
        self._down_indicator.visible = (
            abs(self.layout.view_y) < self.layout.content_height - self.layout.height
        )


class LayoutView(arcade.gui.UIView):
    """This view demonstrates the use of layouts."""

    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_PETER_RIVER

        # Create a anchor layout, which can be used to position widgets on screen
        self.anchor = self.add_widget(UIAnchorLayout())

        # Add describing text in center
        text_area = ScrollableTextArea(
            text=DESCRIPTION,
            text_color=arcade.uicolor.WHITE_CLOUDS,
            font_name=("Lato", "proxima-nova", "Helvetica Neue", "Arial", "sans-serif"),
            font_size=12,
            size_hint=(0.5, 0.8),
        )
        self.anchor.add(text_area, anchor_x="center_x", anchor_y="center_y")
        text_area.with_border(color=arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE)
        text_area.with_background(color=arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE.replace(a=125))
        text_area.with_padding(left=5)

        # add a grid layout with the window and grid size and grid position
        self.grid = arcade.gui.UIGridLayout(
            column_count=2,
            row_count=2,
            align_horizontal="left",
        )
        self.grid.with_background(color=arcade.uicolor.GRAY_ASBESTOS)
        self.grid.with_border(color=arcade.uicolor.GRAY_CONCRETE)
        self.grid.with_padding(all=10)
        self.anchor.add(self.grid, anchor_x="left", anchor_y="top", align_x=10, align_y=-10)
        self.grid.add(
            arcade.gui.UILabel(text="Arcade: "),
            row=0,
            column=0,
        )
        self.grid.add(
            arcade.gui.UILabel(text=arcade.VERSION),
            row=0,
            column=1,
        )
        self.grid.add(
            arcade.gui.UILabel(text="Today: "),
            row=1,
            column=0,
        )
        self.grid.add(
            arcade.gui.UILabel(text=datetime.utcnow().isoformat()[:10]),
            row=1,
            column=1,
        )

        # add a horizontal boxlayout with buttons
        h_box = arcade.gui.UIBoxLayout(space_between=20, vertical=False)
        self.anchor.add(h_box, anchor_x="center_x", anchor_y="bottom", align_y=5)

        # Add a button to move the grid layout to top_right
        # top left
        move_left_button = arcade.gui.UIFlatButton(text="Top Left", width=150)
        move_left_button.disabled = True
        h_box.add(move_left_button)

        @move_left_button.event("on_click")
        def _(event):
            self.anchor.remove(self.grid)
            self.anchor.add(self.grid, anchor_x="left", anchor_y="top", align_x=10, align_y=-10)
            move_left_button.disabled = True
            move_right_button.disabled = False

        move_right_button = arcade.gui.UIFlatButton(text="Top Right", width=150)
        h_box.add(move_right_button)

        @move_right_button.event("on_click")
        def _(event):
            self.anchor.remove(self.grid)
            self.anchor.add(self.grid, anchor_x="right", anchor_y="top", align_x=-10, align_y=-10)
            move_right_button.disabled = True
            move_left_button.disabled = False

    def on_draw_before_ui(self):
        # Add draw commands that should be below the UI
        pass

    def on_draw_after_ui(self):
        # Add draw commands that should be on top of the UI (uncommon)
        pass


def main():
    window = arcade.Window(title="GUI Example: Layouts")
    window.show_view(LayoutView())
    window.run()


if __name__ == '__main__':
    main()



"""
Show how to have enemies shoot bullets at regular intervals.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_bullets_periodic
"""
import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprites and Periodic Bullets Example"


class EnemySprite(arcade.Sprite):
    """ Enemy ship class that tracks how long it has been since firing. """

    def __init__(self, image_file, scale, bullet_list, time_between_firing):
        """ Set up the enemy """
        super().__init__(image_file, scale=scale)

        # How long has it been since we last fired?
        self.time_since_last_firing = 0.0

        # How often do we fire?
        self.time_between_firing = time_between_firing

        # When we fire, what list tracks the bullets?
        self.bullet_list = bullet_list

    def update(self, delta_time: float = 1 / 60):
        """ Update this sprite. """

        # Track time since we last fired
        self.time_since_last_firing += delta_time

        # If we are past the firing time, then fire
        if self.time_since_last_firing >= self.time_between_firing:

            # Reset timer
            self.time_since_last_firing = 0

            # Fire the bullet
            bullet = arcade.Sprite(":resources:images/space_shooter/laserBlue01.png")
            bullet.center_x = self.center_x
            bullet.angle = 90
            bullet.top = self.bottom
            bullet.change_y = -2
            self.bullet_list.append(bullet)


class GameView(arcade.View):
    """ Main application class """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.BLACK

        self.player = None
        self.player_list = None
        self.enemy_list = None
        self.bullet_list = None

    def setup(self):
        """ Set up the variables for the game. """

        self.player_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()

        # Add player ship
        self.player = arcade.Sprite(
            ":resources:images/space_shooter/playerShip1_orange.png",
            scale=0.5,
        )
        self.player_list.append(self.player)

        # Add top-left enemy ship
        enemy = EnemySprite(
            ":resources:images/space_shooter/playerShip1_green.png",
            scale=0.5,
            bullet_list=self.bullet_list,
            time_between_firing=2.0,
        )
        enemy.center_x = 120
        enemy.center_y = WINDOW_HEIGHT - enemy.height
        enemy.angle = 180
        self.enemy_list.append(enemy)

        # Add top-right enemy ship
        enemy = EnemySprite(":resources:images/space_shooter/playerShip1_green.png",
                            scale=0.5,
                            bullet_list=self.bullet_list,
                            time_between_firing=1.0)
        enemy.center_x = WINDOW_WIDTH - 120
        enemy.center_y = WINDOW_HEIGHT - enemy.height
        enemy.angle = 180
        self.enemy_list.append(enemy)

    def on_draw(self):
        """Render the screen. """

        self.clear()

        self.enemy_list.draw()
        self.bullet_list.draw()
        self.player_list.draw()

    def on_update(self, delta_time):
        """ All the logic to move, and the game logic goes here. """

        # Call on_update for each enemy in  the list
        self.enemy_list.update(delta_time)

        # Get rid of the bullet when it flies off-screen
        for bullet in self.bullet_list:
            if bullet.top < 0:
                bullet.remove_from_sprite_lists()

        self.bullet_list.update(delta_time)

    def on_mouse_motion(self, x, y, delta_x, delta_y):
        """
        Called whenever the mouse moves.
        """
        self.player.center_x = x
        self.player.center_y = 20


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



import arcade.gl as gl
from arcade.future.background import Background


class BackgroundGroup:
    """
    If you have many backgrounds which you would like to draw together and move
    together this can help.
    The pos of the Background Group is independent of each Background pos.
    The offset of the BackgroundGroup is the same as each background.
    """

    def __init__(self, backgrounds: list[Background] | None = None):
        self._backgrounds: list[Background] = [] if backgrounds is None else backgrounds

        self._pos = (0.0, 0.0)
        self._offset = (0.0, 0.0)

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, value: tuple[float, float]):
        self._pos = value

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = value
        for background in self._backgrounds:
            background.texture.offset = value

    def __getitem__(self, item: int):
        return self._backgrounds[item]

    def __setitem__(self, key: int, value: Background):
        self._backgrounds[key] = value

    def __iter__(self):
        return iter(self._backgrounds)

    def add(self, item: Background):
        if item not in self._backgrounds:
            self._backgrounds.append(item)
            item.texture.offset = self._offset
        else:
            print("WARNING: Background already in group")

    def extend(self, items: list[Background]):
        for item in items:
            self.add(item)

    def draw(self):
        for background in self._backgrounds:
            background.draw(self.pos)

    def add_from_file(
        self,
        tex_src: str,
        pos: tuple[float, float] = (0.0, 0.0),
        size: tuple[int, int] | None = None,
        offset: tuple[float, float] = (0.0, 0.0),
        scale: float = 1.0,
        angle: float = 0.0,
        *,
        filters=(gl.NEAREST, gl.NEAREST),
        color: tuple[int, int, int] | None = None,
        color_norm: tuple[float, float, float] | None = None,
        shader: gl.Program | None = None,
        geometry: gl.Geometry | None = None,
    ):
        background = Background.from_file(
            tex_src,
            pos,
            size,
            offset,
            scale,
            angle,
            filters=filters,
            color=color,
            color_norm=color_norm,
            shader=shader,
            geometry=geometry,
        )
        self.add(background)


class ParallaxGroup:
    """
    The ParallaxBackground holds a list of backgrounds and a list of depths.
    When you change the offset through the ParallaxBackground
    each Background's offset will be set inversely proportional to its depth.
    This creates the effect of Backgrounds with greater depths appearing further away.
    The depth does not affect the positioning of layers at all.
    """

    def __init__(
        self,
        backgrounds: list[Background] | None = None,
        depths: list[float] | None = None,
    ):
        self._backgrounds: list[Background] = [] if backgrounds is None else backgrounds
        self._depths: list[float] = [] if depths is None else depths

        if len(self._backgrounds) != len(self._depths):
            raise ValueError("The number of backgrounds does not equal the number of depth values")

        self._pos = (0.0, 0.0)
        self._offset = (0.0, 0.0)

    @property
    def pos(self) -> tuple[float, float]:
        return self._pos

    @pos.setter
    def pos(self, value: tuple[float, float]):
        self._pos = value

    @property
    def offset(self) -> tuple[float, float]:
        return self._offset

    @offset.setter
    def offset(self, value: tuple[float, float]):
        self._offset = value
        for index, background in enumerate(self._backgrounds):
            depth = self._depths[index]
            background.texture.offset = value[0] / depth, value[1] / depth

    def __getitem__(self, item: int):
        return self._backgrounds[item], self._depths[item]

    def __setitem__(self, key: int, value: Background | float):
        if isinstance(value, float | int):
            self._depths[key] = value
        else:
            self._backgrounds[key] = value

    def add(self, item: Background, depth: float = 1.0):
        if item not in self._backgrounds:
            self._backgrounds.append(item)
            self._depths.append(depth)
        else:
            print("WARNING: Background already in group")

    def remove(self, item: Background):
        index = self._backgrounds.index(item)
        self._backgrounds.remove(item)
        self._depths.pop(index)

    def change_depth(self, item: Background, new_depth: float):
        self._depths[self._backgrounds.index(item)] = new_depth

    def extend(self, items: list[Background], depths: list[float]):
        for index, item in enumerate(items):
            self.add(item, depths[index])

    def draw(self):
        for background in self._backgrounds:
            background.draw(self.pos)

    def add_from_file(
        self,
        tex_src: str,
        pos: tuple[float, float] = (0.0, 0.0),
        size: tuple[int, int] | None = None,
        depth: float = 1,
        offset: tuple[float, float] = (0.0, 0.0),
        scale: float = 1.0,
        angle: float = 0.0,
        *,
        filters=(gl.NEAREST, gl.NEAREST),
        color: tuple[int, int, int] | None = None,
        color_norm: tuple[float, float, float] | None = None,
        shader: gl.Program | None = None,
        geometry: gl.Geometry | None = None,
    ):
        background = Background.from_file(
            tex_src,
            pos,
            size,
            offset,
            scale,
            angle,
            filters=filters,
            color=color,
            color_norm=color_norm,
            shader=shader,
            geometry=geometry,
        )
        self.add(background, depth)



from __future__ import annotations

from collections.abc import Callable, Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING

from pyglet.math import Mat4, Vec2, Vec3

from arcade.camera.data_types import (
    CameraData,
    OrthographicProjectionData,
    PerspectiveProjectionData,
)
from arcade.camera.projection_functions import (
    generate_orthographic_matrix,
    generate_perspective_matrix,
    generate_view_matrix,
    project_orthographic,
    project_perspective,
    unproject_orthographic,
    unproject_perspective,
)
from arcade.types import Point, Point3
from arcade.window_commands import get_window

if TYPE_CHECKING:
    from arcade.application import Window


class _StaticCamera:
    def __init__(
        self,
        view_matrix: Mat4,
        projection_matrix: Mat4,
        viewport: tuple[int, int, int, int] | None = None,
        *,
        project_method: (
            Callable[[Point, tuple[int, int, int, int], Mat4, Mat4], Vec2] | None
        ) = None,
        unproject_method: (
            Callable[[Point, tuple[int, int, int, int], Mat4, Mat4], Vec3] | None
        ) = None,
        window: Window | None = None,
    ):
        self._win: Window = window or get_window()
        self._viewport: tuple[int, int, int, int] = viewport or self._win.ctx.viewport
        self._view = view_matrix
        self._projection = projection_matrix

        self._project_method: Callable[[Point, tuple, Mat4, Mat4], Vec2] | None = project_method
        self._unproject_method: Callable[[Point, tuple, Mat4, Mat4], Vec3] | None = unproject_method

    def use(self):
        self._win.current_camera = self

        self._win.ctx.viewport = self._viewport
        self._win.ctx.projection_matrix = self._projection
        self._win.ctx.view_matrix = self._view

    @contextmanager
    def activate(self) -> Generator[_StaticCamera, None, None]:
        prev = self._win.ctx.current_camera
        try:
            self.use()
            yield self
        finally:
            prev.use()

    def project(self, world_coordinate: Point) -> Vec2:
        """
        Take a Vec2 or Vec3 of coordinates and return the related screen coordinate
        """
        if self._project_method is None:
            raise ValueError("This Static Camera was not provided a project method at creation")

        pos = self._project_method(world_coordinate, self._viewport, self._view, self._projection)
        return pos

    def unproject(self, screen_coordinate: Point) -> Vec3:
        """
        Take in a pixel coordinate from within
        the range of the window size and returns
        the world space coordinates.

        Essentially reverses the effects of the projector.

        Args:
            screen_coordinate: A 2D position in pixels from the bottom left of the screen.
                               This should ALWAYS be in the range of 0.0 - screen size.
        Returns:
            A 3D vector in world space.
        """
        if self._unproject_method is None:
            raise ValueError("This Static Camera was not provided an unproject method at creation")

        return self._unproject_method(
            screen_coordinate, self._viewport, self._view, self._projection
        )


def static_from_orthographic(
    view: CameraData,
    orthographic: OrthographicProjectionData,
    viewport: tuple[int, int, int, int] | None = None,
    *,
    window: Window | None = None,
) -> _StaticCamera:
    """
    Create a static camera from a CameraData and OrthographicProjectionData

    Args:
        view: The view matrix to use
        orthographic: The orthographic projection to use
        viewport: The viewport to use
        window: The window to use
    """
    return _StaticCamera(
        generate_view_matrix(view),
        generate_orthographic_matrix(orthographic, view.zoom),
        viewport,
        window=window,
        project_method=project_orthographic,
        unproject_method=unproject_orthographic,
    )


def static_from_perspective(
    view: CameraData,
    perspective: OrthographicProjectionData,
    viewport: tuple[int, int, int, int] | None = None,
    *,
    window: Window | None = None,
) -> _StaticCamera:
    """
    Create a static camera from a CameraData and PerspectiveProjectionData

    Args:
        view: The view matrix to use
        perspective: The perspective projection to use
        viewport: The viewport to use
        window: The window
    """
    return _StaticCamera(
        generate_view_matrix(view),
        generate_orthographic_matrix(perspective, view.zoom),
        viewport,
        window=window,
        project_method=project_perspective,
        unproject_method=unproject_perspective,
    )


def static_from_raw_orthographic(
    projection: tuple[float, float, float, float],
    near: float = -100.0,
    far: float = 100.0,
    zoom: float = 1.0,
    position: Point3 = (0.0, 0.0, 0.0),
    up: Point3 = (0.0, 1.0, 0.0),
    forward: Point3 = (0.0, 0.0, -1.0),
    viewport: tuple[int, int, int, int] | None = None,
    *,
    window: Window | None = None,
) -> _StaticCamera:
    """
    Create a static camera from raw orthographic data.

    Args:
        projection: The orthographic projection to use
        near: The near plane
        far: The far plane
        zoom: The zoom level
        position: The position of the camera
        up: The up vector
        forward: The forward vector
        viewport: The viewport
        window: The window
    """
    view = generate_view_matrix(CameraData(position, up, forward, zoom))
    proj = generate_orthographic_matrix(
        OrthographicProjectionData(
            projection[0], projection[1], projection[2], projection[3], near, far
        ),
        zoom,
    )
    return _StaticCamera(
        view,
        proj,
        viewport,
        window=window,
        project_method=project_orthographic,
        unproject_method=unproject_orthographic,
    )


def static_from_raw_perspective(
    aspect: float,
    fov: float,
    near: float = -100.0,
    far: float = 100.0,
    zoom: float = 1.0,
    position: Point3 = (0.0, 0.0, 0.0),
    up: Point3 = (0.0, 1.0, 0.0),
    forward: Point3 = (0.0, 0.0, -1.0),
    viewport: tuple[int, int, int, int] | None = None,
    *,
    window: Window | None = None,
) -> _StaticCamera:
    """
    Create a static camera from raw perspective data.

    Args:
        aspect: The aspect ratio
        fov: The field of view
        near: The near plane
        far: The far plane
        zoom: The zoom level
        position: The position of the camera
        up: The up vector
        forward: The forward vector
        viewport: The viewport
        window: The window
    """
    view = generate_view_matrix(CameraData(position, up, forward, zoom))
    proj = generate_perspective_matrix(PerspectiveProjectionData(aspect, fov, near, far), zoom)

    return _StaticCamera(
        view,
        proj,
        viewport,
        window=window,
        project_method=project_perspective,
        unproject_method=unproject_perspective,
    )


def static_from_matrices(
    view: Mat4,
    projection: Mat4,
    viewport: tuple[int, int, int, int] | None,
    *,
    window: Window | None = None,
    project_method: Callable[[Point, tuple[int, int, int, int], Mat4, Mat4], Vec2] | None = None,
    unproject_method: Callable[[Point, tuple[int, int, int, int], Mat4, Mat4], Vec3] | None = None,
) -> _StaticCamera:
    """
    Create a static camera from raw matrices.

    Args:
        view: The view matrix
        projection: The projection matrix
        viewport: The viewport
        window: The window
        project_method: The project method
        unproject_method: The unproject method
    """
    return _StaticCamera(
        view,
        projection,
        viewport,
        window=window,
        project_method=project_method,
        unproject_method=unproject_method,
    )



from pathlib import Path
import shutil
import sys
import pyglet
import arcade
import PIL

EXAMPLE_PATH = Path(__file__).parent.parent.resolve() / "examples"


def execute_from_command_line():
    if len(sys.argv) == 1:
        show_info()
        return

    command = sys.argv[1]
    if command == "startproject":
        start_project(sys.argv[2])
    else:
        print("Unsupported command")


def show_info():
    window = arcade.Window()
    version_str = f"Arcade {arcade.__version__}"
    print()
    print(version_str)
    print("-" * len(version_str))
    print("vendor:", window.ctx.info.VENDOR)
    print("renderer:", window.ctx.info.RENDERER)
    print("version:", window.ctx.gl_version)
    print("python:", sys.version)
    print("platform:", sys.platform)
    print("pyglet version:", pyglet.version)
    print("PIL version:", PIL.__version__)


def start_project(path_str: str):
    path = Path(path_str)
    if path.exists():
        print("File already exists")

    shutil.copy(EXAMPLE_PATH / "starting_template.py", path)
    print("Created", path)



"""
This is for testing geometry shader shapes. Please keep.
"""

import math
import random
import time

from pyglet.math import Mat4

import arcade

SCREEN_WIDTH = 800
SCREEN_HEIGHT = 600
TITLE = "Shape Test"

# [x] draw_line
# [x] draw_lines
# [x] draw_circle_filled
# [x] draw_ellipse_filled
# [x] draw_circle_outline
# [x] draw_ellipse_outline
# [ ] draw_arc_filled
# [ ] draw_arc_outline
# [ ] draw_parabola_filled
# [ ] draw_parabola_outline
# [ ] draw_line_strip
# [x] draw_point
# [x] draw_points
# [ ] draw_polygon_filled
# [ ] draw_polygon_outline
# [x] draw_triangle_filled
# [x] draw_triangle_outline
# [x] draw_rectangle_outline
# [x] draw_xywh_rectangle_outline
# [x] draw_lrbt_rectangle_filled
# [x] draw_xywh_rectangle_filled
# [x] draw_rectangle_filled
# [x] draw_scaled_texture_rectangle
# [x] draw_texture_rectangle
# [x] draw_lrwh_rectangle_textured

# --- Buffered
# create_line
# create_line_generic_with_colors
# create_line_generic
# create_line_strip
# create_line_loop
# create_lines
# create_lines_with_colors
# create_polygon
# create_rectangle_filled
# create_rectangle_outline
# create_rectangle
# create_rectangle_filled_with_colors
# create_rectangles_filled_with_colors
# create_triangles_filled_with_colors
# create_ellipse_filled
# create_ellipse_outline
# create_ellipse
# create_ellipse_filled_with_colors


def random_pos():
    return random.randrange(0, SCREEN_WIDTH), random.randrange(0, SCREEN_HEIGHT)


def random_color(alpha=127):
    return random.randint(0, 255), random.randint(0, 255), random.randint(0, 255), alpha


def random_radius(start=5, end=25):
    return random.randrange(start, end)


class GameWindow(arcade.Window):
    def __init__(self, width, height, title):
        super().__init__(width, height, title, antialiasing=True, resizable=True)
        self.set_vsync(False)

        # Single lines
        self.single_lines_calls = [
            (*random_pos(), *random_pos(), random_color()) for _ in range(600)
        ]
        # Line list
        self.line_list = [
            (random.randrange(0, SCREEN_WIDTH), random.randrange(0, SCREEN_HEIGHT))
            for _ in range(2 * 10000)
        ]

        # Single circle draw calls
        self.single_circle_calls = [
            (*random_pos(), random_radius(), random_color()) for _ in range(200)
        ]

        # line strip
        self.line_strip = [
            (random.randint(0, SCREEN_WIDTH), random.randint(0, SCREEN_HEIGHT)) for _ in range(10)
        ]
        # Random list of points
        self.points = [
            (random.randint(0, SCREEN_WIDTH), random.randint(0, SCREEN_HEIGHT))
            for _ in range(10_000)
        ]

        self.frames = 0
        self.elapsed = 0
        self.execution_time = 0

    def do_draw_line(self):
        for ln in self.single_lines_calls:
            arcade.draw_line(ln[0], ln[1], ln[2], ln[3], ln[4], 10)

    def do_draw_lines(self):
        arcade.draw_lines(self.line_list, (255, 0, 0, 10))

    def draw_line_strip(self):
        arcade.draw_line_strip(self.line_strip, arcade.color.WHITE, 20.0)

    def do_draw_circle_filled(self):
        for c in self.single_circle_calls:
            arcade.draw_circle_filled(c[0], c[1], c[2], c[3])

    def do_draw_ellipse_filled(self):
        arcade.draw_ellipse_filled(400, 300, 100, 200, arcade.color.AZURE, self.elapsed * 10)

    def do_draw_circle_outline(self):
        arcade.draw_circle_outline(400, 300, 200, arcade.color.AZURE, 10)

    def do_draw_ellipse_outline(self):
        arcade.draw_ellipse_outline(400, 300, 230, 100, arcade.color.AZURE, 10, 45)

    def do_draw_rectangle(self):
        # 0.1 : 1600
        for x in range(0, SCREEN_WIDTH, 20):
            for y in range(0, SCREEN_HEIGHT, 20):
                arcade.draw_rectangle_filled(x + 10, y + 8, 10, 10, arcade.color.AZURE)

    def do_draw_arc_filled(self):
        arcade.draw_arc_filled(
            400,
            300,
            200,
            200,
            arcade.color.AZURE,
            30.0 - math.sin(self.elapsed) * 20.0,
            340.0 + math.sin(self.elapsed) * 20.0,
            0,
        )

    def do_draw_point(self):
        for x in range(0, SCREEN_WIDTH, 20):
            for y in range(0, SCREEN_HEIGHT, 20):
                arcade.draw_point(x, y, arcade.color.WHITE, 1)

    def do_draw_points(self):
        arcade.draw_points(self.points, arcade.color.WILLPOWER_ORANGE, 1.0)

    def do_draw_rectangle_outline(self):
        for x in range(0, SCREEN_WIDTH, 20):
            for y in range(0, SCREEN_HEIGHT, 20):
                arcade.draw_rectangle_outline(
                    x, y, 10, 10, arcade.color.AERO_BLUE, 5, self.elapsed * 100
                )

    def do_draw_xywh_rectangle_outline(self):
        for x in range(0, SCREEN_WIDTH, 20):
            for y in range(0, SCREEN_HEIGHT, 20):
                arcade.draw_xywh_rectangle_outline(x, y, 10, 10, arcade.color.AERO_BLUE, 5)

    def do_draw_triangle_outline(self):
        for x in range(0, SCREEN_WIDTH, 20):
            for y in range(0, SCREEN_HEIGHT, 20):
                arcade.draw_triangle_outline(
                    x, y, x + 10, y, x + 5, y + 10, arcade.color.AERO_BLUE, 2
                )

    def do_draw_triangle_filled(self):
        for x in range(0, SCREEN_WIDTH, 20):
            for y in range(0, SCREEN_HEIGHT, 20):
                arcade.draw_triangle_filled(x, y, x + 10, y, x + 5, y + 10, arcade.color.AERO_BLUE)

    def do_draw_polygon_outline(self):
        arcade.draw_polygon_outline(self.points, arcade.color.AERO_BLUE)

    def on_draw(self):
        self.clear()

        start = time.time()

        # Toggle what to test here
        # self.do_draw_line()
        # self.do_draw_lines()
        # self.draw_line_strip()
        # self.do_draw_circle_filled()
        # self.do_draw_ellipse_filled()
        # self.do_draw_circle_outline()
        # self.do_draw_ellipse_outline()
        # self.do_draw_rectangle()
        # self.do_draw_arc_filled()
        # self.do_draw_point()
        # self.do_draw_points()
        # self.do_draw_rectangle_outline()  # 0.33
        # self.do_draw_xywh_rectangle_outline()
        # self.do_draw_triangle_outline()  # 0.42
        # self.do_draw_triangle_filled()  # 0.26
        self.do_draw_polygon_outline()

        self.ctx.finish()
        self.execution_time += time.time() - start
        self.frames += 1

        if self.execution_time > 1.0 and self.frames > 0:
            print(
                f"frames {self.frames}, "
                f"execution time {round(self.execution_time, 3)}, "
                f"frame time {round(self.execution_time / self.frames, 3)}"
            )
            self.execution_time = 0
            self.frames = 0

    def on_update(self, dt):
        self.elapsed += dt

    def on_resize(self, width: float, height: float):
        w, h = self.get_framebuffer_size()
        self.ctx.viewport = 0, 0, w, h
        self.ctx.projection_matrix = Mat4.orthogonal_projection(0, 800, 0, 600, -100, 100)


if __name__ == "__main__":
    window = GameWindow(SCREEN_WIDTH, SCREEN_HEIGHT, TITLE)
    arcade.run()



"""
Simple demo showing runtime image replacement in texture atlas directly.
The right solution would normally be to replace the sprite with a new
one, but in instances were images are created manually by code and changing
over time at (not too frequently) we can update the underlying atlas directly.
"""

from itertools import cycle

import arcade

TEXTURE_PATHS = [
    ":resources:images/cards/cardClubs2.png",
    ":resources:images/cards/cardClubs3.png",
    ":resources:images/cards/cardClubs4.png",
    ":resources:images/cards/cardClubs5.png",
    ":resources:images/cards/cardClubs6.png",
    ":resources:images/cards/cardClubs7.png",
    ":resources:images/cards/cardClubs8.png",
    ":resources:images/cards/cardClubs9.png",
    ":resources:images/cards/cardClubsA.png",
    ":resources:images/cards/cardClubsJ.png",
    ":resources:images/cards/cardClubsK.png",
    ":resources:images/cards/cardClubsQ.png",
    ":resources:images/cards/cardDiamonds10.png",
    ":resources:images/cards/cardDiamonds2.png",
    ":resources:images/cards/cardDiamonds3.png",
    ":resources:images/cards/cardDiamonds4.png",
    ":resources:images/cards/cardDiamonds5.png",
    ":resources:images/cards/cardDiamonds6.png",
    ":resources:images/cards/cardDiamonds7.png",
    ":resources:images/cards/cardDiamonds8.png",
    ":resources:images/cards/cardDiamonds9.png",
    ":resources:images/cards/cardDiamondsA.png",
    ":resources:images/cards/cardDiamondsJ.png",
    ":resources:images/cards/cardDiamondsK.png",
    ":resources:images/cards/cardDiamondsQ.png",
    ":resources:images/cards/cardHearts10.png",
    ":resources:images/cards/cardHearts2.png",
    ":resources:images/cards/cardHearts3.png",
    ":resources:images/cards/cardHearts4.png",
    ":resources:images/cards/cardHearts5.png",
    ":resources:images/cards/cardHearts6.png",
    ":resources:images/cards/cardHearts7.png",
    ":resources:images/cards/cardHearts8.png",
    ":resources:images/cards/cardHearts9.png",
    ":resources:images/cards/cardHeartsA.png",
    ":resources:images/cards/cardHeartsJ.png",
    ":resources:images/cards/cardHeartsK.png",
    ":resources:images/cards/cardHeartsQ.png",
    ":resources:images/cards/cardJoker.png",
    ":resources:images/cards/cardSpades10.png",
    ":resources:images/cards/cardSpades2.png",
    ":resources:images/cards/cardSpades3.png",
    ":resources:images/cards/cardSpades4.png",
    ":resources:images/cards/cardSpades5.png",
    ":resources:images/cards/cardSpades6.png",
    ":resources:images/cards/cardSpades7.png",
    ":resources:images/cards/cardSpades8.png",
    ":resources:images/cards/cardSpades9.png",
    ":resources:images/cards/cardSpadesA.png",
    ":resources:images/cards/cardSpadesJ.png",
    ":resources:images/cards/cardSpadesK.png",
    ":resources:images/cards/cardSpadesQ.png",
]


class AtlasReplaceImage(arcade.Window):
    def __init__(self):
        super().__init__(800, 600, "Replacing images in atlas")
        self.sprite_1 = arcade.Sprite(TEXTURE_PATHS[-1], center_x=200, center_y=300)
        self.sprite_2 = arcade.Sprite(TEXTURE_PATHS[-2], center_x=400, center_y=300)
        self.sprite_3 = arcade.Sprite(TEXTURE_PATHS[-3], center_x=600, center_y=300)
        self.spritelist = arcade.SpriteList()
        self.spritelist.extend([self.sprite_1, self.sprite_2, self.sprite_3])
        # Timer for texture change
        self.elapsed_time = 0
        self.texture_cycle = cycle(arcade.load_texture(p) for p in TEXTURE_PATHS)

    def on_draw(self):
        self.spritelist.draw()

    def on_update(self, delta_time: float):
        self.elapsed_time += delta_time
        # Change textures
        if self.elapsed_time > 1.0:
            # Replace the internal images. They all have the same size (required)
            self.sprite_1.texture.image = next(self.texture_cycle).image
            self.sprite_2.texture.image = next(self.texture_cycle).image
            self.sprite_3.texture.image = next(self.texture_cycle).image
            # Update the internal texture image in the atlas
            self.spritelist.atlas.update_texture_image(self.sprite_1.texture)
            self.spritelist.atlas.update_texture_image(self.sprite_2.texture)
            self.spritelist.atlas.update_texture_image(self.sprite_3.texture)
            # Reset timer
            self.elapsed_time = 0


window = AtlasReplaceImage()
arcade.run()



"""Creating a hidden password field

This example demonstrates how to create a custom text input
which hides the contents behind a custom character, as often
required for login screens.

Due to a bug in the current version of pyglet, the example uses ENTER to switch
fields instead of TAB. This will be fixed in future versions.
(https://github.com/pyglet/pyglet/issues/1197)

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.exp_hidden_password
"""

import arcade
from arcade.gui import UIInputText, UIOnClickEvent, UIView
from arcade.gui.experimental.password_input import UIPasswordInput
from arcade.gui.widgets.buttons import UIFlatButton
from arcade.gui.widgets.layout import UIGridLayout, UIAnchorLayout
from arcade.gui.widgets.text import UILabel
from arcade import resources

# Load kenny fonts shipped with arcade
resources.load_kenney_fonts()


class MyView(UIView):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_BELIZE_HOLE

        grid = UIGridLayout(
            size_hint=(0, 0),  # wrap children
            row_count=5,  # title | user, pw | login button
            column_count=2,  # label and input field
            vertical_spacing=10,
            horizontal_spacing=5,
        )
        grid.with_padding(all=50)
        grid.with_background(color=arcade.uicolor.GREEN_GREEN_SEA)

        title = grid.add(
            UILabel(text="Login", width=150, font_size=20, font_name="Kenney Future"),
            column=0,
            row=0,
            column_span=2,
        )
        title.with_padding(bottom=20)

        grid.add(UILabel(text="Username:", width=80, font_name="Kenney Future"), column=0, row=1)
        self.username_input = grid.add(
            UIInputText(width=150, font_name="Kenney Future"), column=1, row=1
        )

        grid.add(UILabel(text="Password:", width=80, font_name="Kenney Future"), column=0, row=2)
        self.password_input = grid.add(
            UIPasswordInput(width=150, font_name="Kenney Future"), column=1, row=2
        )
        self.password_input.with_background(color=arcade.uicolor.GREEN_GREEN_SEA)
        # set background to prevent full render on blinking caret

        self.login_button = grid.add(
            UIFlatButton(text="Login", height=30, width=150, size_hint=(1, None)),
            column=0,
            row=3,
            column_span=2,
        )
        self.login_button.on_click = self.on_login

        # add warning label
        self.warning_label = grid.add(
            UILabel(
                text="Use 'TAB' to switch fields, then enter to login",
                width=150,
                font_size=10,
                font_name="Kenney Future",
            ),
            column=0,
            row=4,
            column_span=2,
        )

        anchor = UIAnchorLayout()  # to center grid on screen
        anchor.add(grid)

        self.add_widget(anchor)

        # activate username input field
        self.username_input.activate()

    def on_key_press(self, symbol: int, modifiers: int) -> bool | None:
        # if username field active, switch fields with enter
        if self.username_input.active:
            if symbol == arcade.key.TAB:
                self.username_input.deactivate()
                self.password_input.activate()
                return True
            elif symbol == arcade.key.ENTER:
                self.username_input.deactivate()
                self.on_login(None)
                return True
        # if password field active, login with enter
        elif self.password_input.active:
            if symbol == arcade.key.TAB:
                self.username_input.activate()
                self.password_input.deactivate()
                return True
            elif symbol == arcade.key.ENTER:
                self.password_input.deactivate()
                self.on_login(None)
                return True
        return False

    def on_login(self, event: UIOnClickEvent | None):
        username = self.username_input.text.strip()
        password = self.password_input.text.strip()
        print(f"User logged in with: {username} {password}")


def main():
    window = arcade.Window(title="GUI Example: Hidden Password")
    window.show_view(MyView())
    window.run()


if __name__ == "__main__":
    main()



"""
Rendering to texture with a compute shader

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.compute_texture
"""

import arcade
from arcade.gl import geometry

SIZE = 512, 512

COMPUTE_SHADER = """
#version 430

// Describes the minimal amount of work this compute shader
// can execute. 16 x 16 = 256 threads. When we run a compute
// shader we specify how many of these groups we are executing.
// This is called a work group.
layout (local_size_x = 16, local_size_y = 16) in;

// tell OpenGL what format the image is (rgba8).
// Important: The binding needs to match the unit
// Parameter in Texture.bind_to_image.
// Each image needs a different binding unit starting from 0
layout(rgba8, binding=0) uniform writeonly image2D destTex;

uniform float time;

void main() {
    // texel coordinate we are writing to
    ivec2 texelPos = ivec2(gl_GlobalInvocationID.xy);

    // Calculate 1.0 - distance from the center in each work group
    // gl_LocalInvocationID is the current (x, y) invocation in your work group.
    // We make a 16 x 16 square containing the inverse distance from the center
    float local = 1.0 - length(vec2(ivec2(gl_LocalInvocationID.xy) - 8) / 8.0);

    // Wave covering the screen diagonally using the global invocation value.
    // This would pretty much be the pixel location
    float global = sin(
        float(gl_GlobalInvocationID.x + gl_GlobalInvocationID.y) * 0.01 + time
    ) / 2.0 + 0.5;
    imageStore(destTex, texelPos, vec4(local, global, 0.0, 1.0));
}
"""


class App(arcade.Window):

    def __init__(self, *args, **kwargs):
        # We need to specify OpenGL 4.3 when using Compute Shaders
        super().__init__(*SIZE, "Compute Shader", gl_version=(4, 3))
        self.cs = self.ctx.compute_shader(source=COMPUTE_SHADER)
        # In gles the texture needs to be immutable (immutable storage, not contents)
        self.texture = self.ctx.texture(SIZE, components=4, immutable=True)
        self.texture.filter = self.ctx.NEAREST, self.ctx.NEAREST

        # Simple program rendering textured rectangle to the screen
        # using normalized device coordinates (no projection)
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_vert;
            in vec2 in_uv;
            out vec2 uv;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
                uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D image;
            in vec2 uv;
            out vec4 fragColor;

            void main() {
                fragColor = texture(image, uv);
            }
            """,
        )
        self.quad = geometry.quad_2d_fs()

    def on_draw(self):
        self.clear()

        # Calculate invocation parameters
        w, h = self.texture.size
        # Group size is configured to be 16 x 16 in the shader source.
        gw, gh = 16, 16
        # Figure how many groups (16 x 16 tiles) the image consists of
        # This example is a 512 x 512 by default so we will will get
        # 32 x 32 groups/tiles = 1024 work groups executed in total.
        nx, ny = w // gw, h // gh

        # Bind image + run compute shader
        self.texture.bind_to_image(0, write=True)

        # Run the compute shader
        self.cs.run(group_x=nx, group_y=ny)

        # Visualize the generated texture
        self.texture.use(0)
        self.quad.render(self.program)

    def on_update(self, delta_time: float):
        self.cs["time"] = self.time * 10


app = App()
arcade.run()



"""
Sprite Follow Path

This example has enemy sprites follow a set path.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.follow_path
"""

import arcade
import math

# --- Constants ---
SPRITE_SCALING_PLAYER = 0.5
SPRITE_SCALING_ENEMY = 0.5
ENEMY_SPEED = 5.0

WINDOW_WIDTH = 1289
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Follow Path Simple Example"


class Enemy(arcade.Sprite):
    """
    This class represents the Enemy on our screen.
    """

    def __init__(self, image, scale, position_list):
        super().__init__(image, scale=scale)
        self.position_list = position_list
        self.cur_position = 0
        self.speed = ENEMY_SPEED

    def update(self, delta_time: float = 1 / 60):
        """Have a sprite follow a path"""
        # Where are we
        start_x = self.center_x
        start_y = self.center_y

        # Where are we going
        dest_x = self.position_list[self.cur_position][0]
        dest_y = self.position_list[self.cur_position][1]

        # X and Y diff between the two
        x_diff = dest_x - start_x
        y_diff = dest_y - start_y

        # Calculate angle to get there
        angle = math.atan2(y_diff, x_diff)

        # How far are we?
        distance = math.sqrt((self.center_x - dest_x) ** 2 + (self.center_y - dest_y) ** 2)

        # How fast should we go? If we are close to our destination,
        # lower our speed so we don't overshoot.
        speed = min(self.speed, distance)

        # Calculate vector to travel
        change_x = math.cos(angle) * speed
        change_y = math.sin(angle) * speed

        # Update our location
        self.center_x += change_x
        self.center_y += change_y

        # How far are we?
        distance = math.sqrt((self.center_x - dest_x) ** 2 + (self.center_y - dest_y) ** 2)

        # If we are there, head to the next point.
        if distance <= self.speed:
            self.cur_position += 1

            # Reached the end of the list, start over.
            if self.cur_position >= len(self.position_list):
                self.cur_position = 0


class GameView(arcade.View):

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = None
        self.enemy_list = None

        # Set up the player info
        self.player_sprite = None
        self.score = 0

        # Don't show the mouse cursor
        self.window.set_mouse_visible(False)

        self.background_color = arcade.color.AMAZON

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.enemy_list = arcade.SpriteList()

        # Score
        self.score = 0

        # Set up the player
        # Character image from kenney.nl
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING_PLAYER)
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 50
        self.player_list.append(self.player_sprite)

        # List of points the enemy will travel too.
        position_list = [
            [50, 50],
            [self.width - 50, 50],
            [self.width - 50, self.height - 50],
            [50, self.height - 50],
        ]
        # Spawn a few enemies
        self.spawn_enemy((50, 50), position_list)
        self.spawn_enemy((50, 250), position_list)
        self.spawn_enemy((50, 450), position_list)
        self.spawn_enemy((50, 650), position_list)

    def spawn_enemy(self, position, position_list):
        # Spawn a new enemy
        enemy = Enemy(":resources:images/animated_characters/robot/robot_idle.png",
                      SPRITE_SCALING_ENEMY,
                      position_list)
        # Set initial location of the enemy at the first point
        enemy.position = position
        self.enemy_list.append(enemy)

    def on_draw(self):
        """ Draw everything """
        self.clear()
        self.enemy_list.draw()
        self.player_list.draw()

    def on_mouse_motion(self, x, y, dx, dy):
        """ Handle Mouse Motion """
        # Move the center of the player sprite to match the mouse x, y
        self.player_sprite.center_x = x
        self.player_sprite.center_y = y

    def on_update(self, delta_time):
        """ Movement and game logic """
        self.enemy_list.update()

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.ESCAPE:
            self.window.close()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from PIL.Image import Image

from arcade.types import Point, Point2List

from .base import HitBoxAlgorithm


class SimpleHitBoxAlgorithm(HitBoxAlgorithm):
    """
    Simple hit box algorithm. This algorithm attempts to trim out transparent pixels
    from an image to create a hit box.
    """

    def calculate(self, image: Image, **kwargs) -> Point2List:
        """
        Given an RGBA image, this returns points that make up a hit box around it. Attempts
        to trim out transparent pixels.

        Args:
            image: Image get hit box from.
        """
        if image.mode != "RGBA":
            raise ValueError("Image mode is not RGBA. image.convert('RGBA') is needed.")

        # Convert the image into one channel alpha since we don't care about RGB values
        image = image.getchannel("A")
        bbox = image.getbbox()
        # If there is no bounding box the image is empty
        if bbox is None:
            return self.create_bounding_box(image)

        left_border, top_border, right_border, bottom_border = bbox
        right_border -= 1
        bottom_border -= 1

        def _check_corner_offset(
            start_x: int, start_y: int, x_direction: int, y_direction: int
        ) -> int:
            bad = False
            offset = 0
            while not bad:
                y = start_y + (offset * y_direction)
                x = start_x
                for _ in range(offset + 1):
                    alpha = image.getpixel((x, y))
                    # print(f"({x}, {y}) = {my_pixel} | ", end="")
                    if alpha != 0:
                        bad = True
                        break
                    y -= y_direction
                    x += x_direction
                # print(f" - {bad}")
                if not bad:
                    offset += 1
            # print(f"offset: {offset}")
            return offset

        def _r(point: tuple[float, float], height: int, width: int) -> Point:
            return point[0] - width / 2, (height - point[1]) - height / 2

        top_left_corner_offset = _check_corner_offset(left_border, top_border, 1, 1)
        top_right_corner_offset = _check_corner_offset(right_border, top_border, -1, 1)
        bottom_left_corner_offset = _check_corner_offset(left_border, bottom_border, 1, -1)
        bottom_right_corner_offset = _check_corner_offset(right_border, bottom_border, -1, -1)

        p1 = left_border + top_left_corner_offset, top_border
        p2 = (right_border + 1) - top_right_corner_offset, top_border
        p3 = (right_border + 1), top_border + top_right_corner_offset
        p4 = (right_border + 1), (bottom_border + 1) - bottom_right_corner_offset
        p5 = (right_border + 1) - bottom_right_corner_offset, (bottom_border + 1)
        p6 = left_border + bottom_left_corner_offset, (bottom_border + 1)
        p7 = left_border, (bottom_border + 1) - bottom_left_corner_offset
        p8 = left_border, top_border + top_left_corner_offset

        result = []

        h = image.height
        w = image.width

        result.append(_r(p7, h, w))
        if bottom_left_corner_offset:
            result.append(_r(p6, h, w))

        result.append(_r(p5, h, w))
        if bottom_right_corner_offset:
            result.append(_r(p4, h, w))

        result.append(_r(p3, h, w))
        if top_right_corner_offset:
            result.append(_r(p2, h, w))

        result.append(_r(p1, h, w))
        if top_left_corner_offset:
            result.append(_r(p8, h, w))

        # Remove duplicates
        return tuple(dict.fromkeys(result))  # type: ignore



"""
A 3D sphere example.

We're showing how to use the geometry.sphere() function to create a sphere
and how different context flags affects the rendering of a 3d object.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.3d_sphere
"""

import arcade
from arcade.math import clamp
from arcade.gl import geometry
from pyglet.math import Mat4, Vec3
from pyglet.graphics import Batch


class Sphere3D(arcade.Window):
    def __init__(self, width, height, title):
        super().__init__(width, height, title, resizable=True)
        self.sphere = geometry.sphere(1.0, 32, 32, uvs=False)
        # Simple color lighting program
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform WindowBlock {
                mat4 projection;
                mat4 view;
            } window;

            in vec3 in_position;
            in vec3 in_normal;

            out vec3 normal;
            out vec3 pos;

            void main() {
                vec4 p = window.view * vec4(in_position, 1.0);
                gl_Position = window.projection * p;
                mat3 m_normal = transpose(inverse(mat3(window.view)));
                normal = m_normal * in_normal;
                pos = p.xyz;
            }
            """,
            fragment_shader="""
            #version 330

            out vec4 fragColor;

            in vec3 normal;
            in vec3 pos;

            void main()
            {
                float l = dot(normalize(-pos), normalize(normal));
                // Draw front and back face differently
                if (l < 0.0) {
                    l = abs(l);
                    fragColor = vec4(0.75, 0.0, 0.0, 1.0) * (0.25 + abs(l) * 0.75);
                } else {
                    fragColor = vec4(1.0) * (0.25 + abs(l) * 0.75);
                }
            }
            """,
        )
        self.on_resize(*self.get_size())

        self.text_batch = Batch()
        self.text_cull = arcade.Text(
            "F2: Toggle cull face (True)",
            x=10,
            y=10,
            font_size=15,
            color=arcade.color.WHITE,
            batch=self.text_batch,
        )
        self.text_depth = arcade.Text(
            "F1: Toggle depth test (True)",
            x=10,
            y=30,
            font_size=15,
            color=arcade.color.WHITE,
            batch=self.text_batch,
        )
        self.text_wireframe = arcade.Text(
            "SPACE: Toggle wireframe (False)",
            x=10,
            y=50,
            font_size=15,
            color=arcade.color.WHITE,
            batch=self.text_batch,
        )
        self.text_fs = arcade.Text(
            "F: Toggle fullscreen (False)",
            x=10,
            y=70,
            font_size=15,
            color=arcade.color.WHITE,
            batch=self.text_batch,
        )
        self.text_vert_count = arcade.Text(
            "Use mouse wheel to add/remove vertices",
            x=10,
            y=90,
            font_size=15,
            color=arcade.color.WHITE,
            batch=self.text_batch,
        )
        self.text_rotate = arcade.Text(
            "Drag mouse to rotate object",
            x=10,
            y=110,
            font_size=15,
            color=arcade.color.WHITE,
            batch=self.text_batch,
        )

        self.set_vsync(True)

        self.rot_x = 0
        self.rot_y = 0
        self.wireframe = True
        self.vert_count = 0.5
        self.drag_time = None
        self.flags = set([self.ctx.DEPTH_TEST])

    def on_draw(self):
        self.clear()
        self.ctx.enable_only(*self.flags)
        self.ctx.wireframe = self.wireframe
        time = self.drag_time or self.time

        # Position and rotate the sphere
        translate = Mat4.from_translation(Vec3(0, 0, -2.5))
        rx = Mat4.from_rotation(time + self.rot_x, Vec3(0, 1, 0))
        ry = Mat4.from_rotation(time + self.rot_y, Vec3(1, 0, 0))
        # Set matrices and draw
        self.view = translate @ rx @ ry
        self.projection = Mat4.perspective_projection(self.aspect_ratio, 0.1, 100, fov=60)
        self.sphere.render(self.program, vertices=int(self.sphere.num_vertices * self.vert_count))

        # Switch to 2D mode when drawing text
        self.projection = Mat4.orthogonal_projection(0, self.width, 0, self.height, -1, 1)
        self.ctx.disable(self.ctx.DEPTH_TEST, self.ctx.CULL_FACE)
        self.ctx.wireframe = False
        self.view = Mat4()

        with self.ctx.enabled_only():
            self.text_batch.draw()

    def on_key_press(self, key, modifiers):
        if key == arcade.key.ESCAPE:
            self.close()

        elif key == arcade.key.F:
            self.set_fullscreen(not self.fullscreen)
            self.text_fs.text = f"F: Toggle fullscreen ({self.fullscreen})"

        elif key == arcade.key.SPACE:
            self.wireframe = not self.wireframe
            self.set_vsync(True)
            self.text_wireframe.text = f"SPACE: Toggle wireframe ({self.wireframe})"

        elif key == arcade.key.F1:
            if self.ctx.DEPTH_TEST in self.flags:
                self.flags.remove(self.ctx.DEPTH_TEST)
            else:
                self.flags.add(self.ctx.DEPTH_TEST)
            self.text_depth.text = f"F1: Toggle depth test ({self.ctx.DEPTH_TEST in self.flags})"

        elif key == arcade.key.F2:
            if self.ctx.CULL_FACE in self.flags:
                self.flags.remove(self.ctx.CULL_FACE)
            else:
                self.flags.add(self.ctx.CULL_FACE)
            self.text_cull.text = f"F2: Toggle cull face ({self.ctx.CULL_FACE in self.flags})"

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        self.drag_time = self.time
        self.rot_x += dx / 100
        self.rot_y -= dy / 100

    def on_mouse_release(self, x, y, button, modifiers):
        self.drag_time = None

    def on_mouse_scroll(self, x: int, y: int, scroll_x: float, scroll_y: float):
        self.vert_count = clamp(self.vert_count + scroll_y / 500, 0.0, 1.0)


if __name__ == "__main__":
    window = Sphere3D(1280, 720, "3D Cube")
    window.set_vsync(True)
    arcade.run()



import bisect
import logging
import math

from arcade import Texture
from arcade.types import Point2

from .enums import (
    FACE_DOWN,
    FACE_LEFT,
    FACE_RIGHT,
    FACE_UP,
)
from .sprite import Sprite

logger = logging.getLogger("arcade")


class TextureKeyframe:
    """
    Keyframe for texture animations.

    Args:
        texture:
            Texture to display for this keyframe.
        duration:
            Duration in milliseconds to display this keyframe.
        tile_id:
            Tile ID for this keyframe (only used for tiled maps).
            This can be ignored when not using tiled maps.
    """

    __slots__ = ("texture", "duration", "tile_id")

    def __init__(self, texture: Texture, duration: int = 100, tile_id: int | None = 0, **kwargs):
        #: The texture to display for this keyframe.
        self.texture = texture
        #: Duration in milliseconds to display this keyframe.
        self.duration = duration
        #: Tile ID for this keyframe (only used for tiled maps)
        self.tile_id = tile_id


class TextureAnimation:
    """
    Animation class that holds a list of keyframes.
    The animation should not store any state related to the current time
    so it can be shared between multiple sprites.

    Args:
        keyframes:
            List of keyframes for the animation.
    """

    __slots__ = ("_keyframes", "_duration_ms", "_timeline")

    def __init__(self, keyframes: list[TextureKeyframe]):
        self._keyframes = keyframes
        self._duration_ms = 0
        self._timeline: list[int] = self._create_timeline(self._keyframes)

    @property
    def keyframes(self) -> tuple[TextureKeyframe, ...]:
        """
        A tuple of keyframes in the animation.
        Keyframes should not be modified directly.
        """
        return tuple(self._keyframes)

    @property
    def duration_seconds(self) -> float:
        """Total duration of the animation in seconds."""
        return self._duration_ms / 1000

    @property
    def duration_ms(self) -> int:
        """Total duration of the animation in milliseconds."""
        return self._duration_ms

    @property
    def num_frames(self) -> int:
        """Number of frames in the animation."""
        return len(self._keyframes)

    def _create_timeline(self, keyframes: list[TextureKeyframe]) -> list[int]:
        """
        Create a timeline of the animation.

        This is a list of timestamps for each frame in seconds.
        """
        timeline: list[int] = []
        current_time_ms = 0
        for frame in keyframes:
            timeline.append(current_time_ms)
            current_time_ms += frame.duration

        self._duration_ms = current_time_ms
        return timeline

    def get_keyframe(self, time: float, loop: bool = True) -> tuple[int, TextureKeyframe]:
        """
        Get the frame at a given time.

        Args:
            time:
                Time in seconds.
            loop:
                If the animation should loop.
        Returns:
            Tuple of frame index and keyframe.
        """
        if loop:
            time_ms = int(time * 1000) % self._duration_ms
        else:
            time_ms = int(time * 1000)

        # Find the right insertion point for the time: O(log n)
        index = bisect.bisect_right(self._timeline, time_ms) - 1
        index = max(0, min(index, len(self._keyframes) - 1))
        return index, self._keyframes[index]

    def __len__(self) -> int:
        return len(self._keyframes)


# Old name: AnimatedTimeBasedSprite
class TextureAnimationSprite(Sprite):
    """
    Animated sprite based on keyframes.

    Primarily used internally by tilemaps to animate tiles.

    Args:
        path_or_texture:
            Path to the image file, or a Texture object.
        center_x:
            Initial x position of the sprite.
        center_y:
            Initial y position of the sprite.
        scale:
            Initial scale of the sprite.
    """

    def __init__(
        self,
        center_x: float = 0.0,
        center_y: float = 0.0,
        scale: float = 1.0,
        animation: TextureAnimation | None = None,
        **kwargs,
    ):
        super().__init__(
            scale=scale,
            center_x=center_x,
            center_y=center_y,
        )
        self._time = 0.0
        self._animation: TextureAnimation | None = None
        if animation:
            self.animation = animation
        self._current_keyframe_index = 0

    @property
    def time(self) -> float:
        """
        Get or set the current time of the animation in seconds.
        """
        return self._time

    @time.setter
    def time(self, value: float) -> None:
        self._time = value

    @property
    def animation(self) -> TextureAnimation:
        """
        Animation object for this sprite.
        """
        if self._animation is None:
            raise RuntimeError("No animation set for this sprite.")
        return self._animation

    @animation.setter
    def animation(self, value: TextureAnimation) -> None:
        """
        Set the animation for this sprite.

        Args:
            value: Animation to set.
        """
        self._animation = value
        # TODO: Forcing the first frame here might not be the best idea.
        self.texture = value._keyframes[0].texture
        self.sync_hit_box_to_texture()

    def update_animation(self, delta_time: float = 1 / 60, **kwargs) -> None:
        """
        Logic for updating the animation.

        Args:
            delta_time: Time since last update.
        """
        if self._animation is None:
            raise RuntimeError("No animation set for this sprite.")

        self.time += delta_time
        index, keyframe = self._animation.get_keyframe(self.time)
        if index != self._current_keyframe_index:
            self._current_keyframe_index = index
            self.texture = keyframe.texture


class AnimatedWalkingSprite(Sprite):
    """
    Deprecated Sprite for platformer games that supports walking animations.
    Make sure to call update_animation after loading the animations so the
    initial texture can be set. Or manually set it.

    It is highly recommended you create your own version of this class rather than
    try to use this pre-packaged one.

    For an example, see this section of the platformer tutorial:
    :ref:`platformer_part_twelve`.

    Args:
        scale:
            Initial scale of the sprite.
        center_x:
            Initial x position of the sprite.
        center_y:
            Initial y position of the sprite.
    """

    def __init__(
        self,
        scale: float | Point2 = 1.0,
        center_x: float = 0.0,
        center_y: float = 0.0,
        **kwargs,
    ):
        super().__init__(
            None,
            scale=scale,
            center_x=center_x,
            center_y=center_y,
        )
        self.state = FACE_RIGHT
        self.stand_right_textures: list[Texture] = []
        self.stand_left_textures: list[Texture] = []
        self.walk_left_textures: list[Texture] = []
        self.walk_right_textures: list[Texture] = []
        self.walk_up_textures: list[Texture] = []
        self.walk_down_textures: list[Texture] = []
        self.cur_texture_index = 0
        self.texture_change_distance = 20
        self.last_texture_change_center_x: float = 0.0
        self.last_texture_change_center_y: float = 0.0

    def update_animation(self, delta_time: float = 1 / 60) -> None:
        """
        Logic for texture animation.

        Args:
            delta_time: Time since last update.
        """
        x1 = self.center_x
        x2 = self.last_texture_change_center_x
        y1 = self.center_y
        y2 = self.last_texture_change_center_y
        distance = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        texture_list: list[Texture] = []

        change_direction = False
        if (
            self.change_x > 0
            and self.change_y == 0
            and self.state != FACE_RIGHT
            and len(self.walk_right_textures) > 0
        ):
            self.state = FACE_RIGHT
            change_direction = True
        elif (
            self.change_x < 0
            and self.change_y == 0
            and self.state != FACE_LEFT
            and len(self.walk_left_textures) > 0
        ):
            self.state = FACE_LEFT
            change_direction = True
        elif (
            self.change_y < 0
            and self.change_x == 0
            and self.state != FACE_DOWN
            and len(self.walk_down_textures) > 0
        ):
            self.state = FACE_DOWN
            change_direction = True
        elif (
            self.change_y > 0
            and self.change_x == 0
            and self.state != FACE_UP
            and len(self.walk_up_textures) > 0
        ):
            self.state = FACE_UP
            change_direction = True

        if self.change_x == 0 and self.change_y == 0:
            if self.state == FACE_LEFT:
                self.texture = self.stand_left_textures[0]
            elif self.state == FACE_RIGHT:
                self.texture = self.stand_right_textures[0]
            elif self.state == FACE_UP:
                self.texture = self.walk_up_textures[0]
            elif self.state == FACE_DOWN:
                self.texture = self.walk_down_textures[0]

        elif change_direction or distance >= self.texture_change_distance:
            self.last_texture_change_center_x = self.center_x
            self.last_texture_change_center_y = self.center_y

            if self.state == FACE_LEFT:
                texture_list = self.walk_left_textures
                if texture_list is None or len(texture_list) == 0:
                    raise RuntimeError(
                        "update_animation was called on a sprite that doesn't have a "
                        "list of walk left textures."
                    )
            elif self.state == FACE_RIGHT:
                texture_list = self.walk_right_textures
                if texture_list is None or len(texture_list) == 0:
                    raise RuntimeError(
                        "update_animation was called on a sprite that doesn't have a list of "
                        "walk right textures."
                    )
            elif self.state == FACE_UP:
                texture_list = self.walk_up_textures
                if texture_list is None or len(texture_list) == 0:
                    raise RuntimeError(
                        "update_animation was called on a sprite that doesn't have a list of "
                        "walk up textures."
                    )
            elif self.state == FACE_DOWN:
                texture_list = self.walk_down_textures
                if texture_list is None or len(texture_list) == 0:
                    raise RuntimeError(
                        "update_animation was called on a sprite that doesn't "
                        "have a list of walk down textures."
                    )

            self.cur_texture_index += 1
            if self.cur_texture_index >= len(texture_list):
                self.cur_texture_index = 0

            self.texture = texture_list[self.cur_texture_index]

        if self._texture is None:
            logger.warning("Error, no texture set")
        else:
            self.width = self._texture.width * self.scale_x
            self.height = self._texture.height * self.scale_x



from __future__ import annotations

from math import cos, radians, sin
from typing import Any

from PIL.Image import Image
from typing_extensions import Self

from arcade.types import EMPTY_POINT_LIST, Point2, Point2List

__all__ = ["HitBoxAlgorithm", "HitBox", "RotatableHitBox"]


class HitBoxAlgorithm:
    """
    The base class for hit box algorithms.

    Hit box algorithms are intended to calculate the points which make up
    a hit box for a given :py:class:`~PIL.Image.Image`. However, advanced
    users can also repurpose them for other tasks.
    """

    #: Whether points for this algorithm should be cached
    cache = True

    def __init__(self):
        self._cache_name = self.__class__.__name__

    @property
    def cache_name(self) -> str:
        """
        A string representation of the parameters used to create this algorithm.

        It will be incorporated at the end of the string returned by
        :py:meth:`Texture.create_cache_name <arcade.Texture.create_cache_name>`.
        Subclasses should override this method to return a value which allows
        distinguishing different configurations of a particular hit box
        algorithm.
        """
        return self._cache_name

    def calculate(self, image: Image, **kwargs) -> Point2List:
        """
        Calculate hit box points for a given image.

        .. warning:: This method should not be made into a class method!

                     Although this base class does not take arguments
                     when initialized, subclasses use them to alter how
                     a specific instance handles image data by default.

        Args:
            image:
                The image to calculate hitbox points for
            kwargs:
                keyword arguments
        """
        raise NotImplementedError

    def __call__(self, *args: Any, **kwds: Any) -> Self:
        """
        Shorthand allowing any instance to be used identically to the base type.

        Args:
            args:
                The same positional arguments as `__init__`
           kwds:
                The same keyword arguments as `__init__`
        Returns:
            A new HitBoxAlgorithm instance
        """
        return self.__class__(*args, **kwds)  # type: ignore

    def create_bounding_box(self, image: Image) -> Point2List:
        """
        Create points for a simple bounding box around an image.
        This is often used as a fallback if a hit box algorithm
        doesn't manage to figure out any reasonable points for
        an image.

        Args:
            image: The image to create a bounding box for.
        """
        size = image.size
        return (
            (-size[0] / 2, -size[1] / 2),
            (size[0] / 2, -size[1] / 2),
            (size[0] / 2, size[1] / 2),
            (-size[0] / 2, size[1] / 2),
        )


class HitBox:
    """
    A basic hit box class supporting scaling.

    It includes support for rescaling as well as shorthand properties
    for boundary values along the X and Y axes. For rotation support,
    use :py:meth:`.create_rotatable` to create an instance of
    :py:class:`RotatableHitBox`.

    Args:
        points:
            The unmodified points bounding the hit box
        position:
            The center around which the points will be offset
        scale:
            The X and Y scaling factors to use when offsetting the points
    """

    def __init__(
        self,
        points: Point2List,
        position: Point2 = (0.0, 0.0),
        scale: Point2 = (1.0, 1.0),
    ):
        self._points = points
        self._position = position
        self._scale = scale

        # This empty tuple will be replaced the first time
        # get_adjusted_points is called
        self._adjusted_points: Point2List = EMPTY_POINT_LIST
        self._adjusted_cache_dirty = True

    @property
    def points(self) -> Point2List:
        """
        The raw, unadjusted points of this hit box.

        These are the points as originally passed before offsetting, scaling,
        and any operations subclasses may perform, such as rotation.
        """
        return self._points

    @property
    def position(self) -> Point2:
        """
        The center point used to offset the final adjusted positions.
        """
        return self._position

    @position.setter
    def position(self, position: Point2):
        self._position = position
        self._adjusted_cache_dirty = True

    # Per Clepto's testing as of around May 2023, these are better
    # left uncached because caching them is somehow slower than what
    # we currently do. Any readers should feel free to retest /
    # investigate further.
    @property
    def left(self) -> float:
        """
        Calculates the leftmost adjusted x position of this hit box
        """
        points = self.get_adjusted_points()
        x_points = [point[0] for point in points]
        return min(x_points)

    @property
    def right(self) -> float:
        """
        Calculates the rightmost adjusted x position of this hit box
        """
        points = self.get_adjusted_points()
        x_points = [point[0] for point in points]
        return max(x_points)

    @property
    def top(self) -> float:
        """
        Calculates the topmost adjusted y position of this hit box
        """
        points = self.get_adjusted_points()
        y_points = [point[1] for point in points]
        return max(y_points)

    @property
    def bottom(self) -> float:
        """
        Calculates the bottommost adjusted y position of this hit box
        """
        points = self.get_adjusted_points()
        y_points = [point[1] for point in points]
        return min(y_points)

    @property
    def scale(self) -> tuple[float, float]:
        """
        The X & Y scaling factors for the points in this hit box.

        These are used to calculate the final adjusted positions of points.
        """
        return self._scale

    @scale.setter
    def scale(self, scale: tuple[float, float]):
        self._scale = scale
        self._adjusted_cache_dirty = True

    def create_rotatable(
        self,
        angle: float = 0.0,
    ) -> RotatableHitBox:
        """
        Create a rotatable instance of this hit box.

        The internal ``PointList`` is transferred directly instead of
        deep copied, so care should be taken if using a mutable internal
        representation.

        Args:
            angle: The angle to rotate points by (0 by default)
        """
        return RotatableHitBox(
            self._points, position=self._position, scale=self._scale, angle=angle
        )

    def get_adjusted_points(self) -> Point2List:
        """
        Return the positions of points, scaled and offset from the center.

        Unlike the boundary helper properties (left, etc), this method will
        only recalculate the values when necessary:

        * The first time this method is called
        * After properties affecting adjusted position were changed
        """
        if not self._adjusted_cache_dirty:
            return self._adjusted_points  # type: ignore

        position_x, position_y = self._position
        scale_x, scale_y = self._scale

        def _adjust_point(point) -> Point2:
            x, y = point

            x *= scale_x
            y *= scale_y

            return (x + position_x, y + position_y)

        self._adjusted_points = [_adjust_point(point) for point in self._points]
        self._adjusted_cache_dirty = False
        return self._adjusted_points


class RotatableHitBox(HitBox):
    """
    A hit box with support for rotation.

    Rotation is separated from the basic hitbox because it is much
    slower than offsetting and scaling.

    Args:
        points:
            The unmodified points bounding the hit box
        position:
            The translation to apply to the points
        angle:
            The angle to rotate the points by
        scale:
            The X and Y scaling factors
    """

    def __init__(
        self,
        points: Point2List,
        *,
        position: tuple[float, float] = (0.0, 0.0),
        angle: float = 0.0,
        scale: Point2 = (1.0, 1.0),
    ):
        super().__init__(points, position=position, scale=scale)
        self._angle: float = angle

    @property
    def angle(self) -> float:
        """
        The angle to rotate the raw points by in degrees
        """
        return self._angle

    @angle.setter
    def angle(self, angle: float):
        self._angle = angle
        self._adjusted_cache_dirty = True

    def get_adjusted_points(self) -> Point2List:
        """
        Return the offset, scaled, & rotated points of this hitbox.

        As with :py:meth:`.HitBox.get_adjusted_points`, this method only
        recalculates the adjusted values when necessary.
        """
        if not self._adjusted_cache_dirty:
            return self._adjusted_points

        rad = radians(-self._angle)
        scale_x, scale_y = self._scale
        position_x, position_y = self._position
        rad_cos = cos(rad)
        rad_sin = sin(rad)

        def _adjust_point(point) -> Point2:
            x, y = point

            x *= scale_x
            y *= scale_y

            if rad:
                rot_x = x * rad_cos - y * rad_sin
                rot_y = x * rad_sin + y * rad_cos
                x = rot_x
                y = rot_y

            return (
                x + position_x,
                y + position_y,
            )

        self._adjusted_points = [_adjust_point(point) for point in self._points]
        self._adjusted_cache_dirty = False
        return self._adjusted_points



"""
Render into a sub-section of a texture atlas
"""

import math

import arcade


class AtlasRenderDemo(arcade.Window):
    def __init__(self):
        super().__init__(1280, 720, "Atlas Render Demo")
        self.atlas = arcade.DefaultTextureAtlas((600, 600))

        self.texture_1 = arcade.Texture.create_empty("render_area_1", size=(256, 256))
        self.texture_2 = arcade.Texture.create_empty("render_area_2", size=(256, 256))
        self.sprite_1 = arcade.Sprite(self.texture_1, center_x=400, center_y=360)
        self.sprite_2 = arcade.Sprite(self.texture_2, center_x=850, center_y=360)

        self.spritelist = arcade.SpriteList(atlas=self.atlas)
        self.spritelist.extend([self.sprite_1, self.sprite_2])

        self.elapsed_time = 0
        self.frame = 0

    def on_draw(self):
        self.clear()
        self.render_into_sprite_texture()
        self.spritelist.draw()

    def render_into_sprite_texture(self):
        # Render shape into texture atlas in the first sprite texture's space
        with self.spritelist.atlas.render_into(self.texture_1) as fbo:
            fbo.clear(color=(255, 0, 0, 255))
            arcade.draw_rect_filled(
                arcade.rect.XYWH(128, 128, 160, 160),
                arcade.color.WHITE,
                self.elapsed_time * 100,
            )

        # Render a shape into the second texture in the atlas
        with self.spritelist.atlas.render_into(self.texture_2) as fbo:
            fbo.clear(color=(0, 255, 0, 255))
            arcade.draw_circle_filled(
                128,
                128,
                80 + math.sin(self.elapsed_time) * 30,
                arcade.color.BLUE,
            )

        self.sprite_1.angle = self.elapsed_time * 66
        self.sprite_2.angle = self.elapsed_time * 56

        self.frame += 1

    def on_update(self, delta_time: float):
        self.elapsed_time += delta_time


window = AtlasRenderDemo()
arcade.run()



import random

import pyglet.clock
from pyglet.graphics import Batch
from pyglet.shapes import Line

import arcade
from arcade.types import RGBA255, Color

__all__ = ["PerfGraph"]


class PerfGraph(arcade.Sprite):
    """
    An auto-updating line chart of FPS or event handler execution times.

    You must use :func:`arcade.enable_timings` to turn on performance
    tracking for the chart to display data.

    Aside from instantiation and updating the chart, this class behaves
    like other :class:`arcade.Sprite` instances. You can use it with
    :class:`SpriteList <arcade.SpriteList>` normally. See
    :ref:`performance_statistics_example` for an example of how to use
    this class.

    Unlike other :class:`Sprite <arcade.Sprite>` instances, this class
    neither loads an :class:`arcade.Texture` nor accepts one as a
    constructor argument. Instead, it creates a new internal
    :class:`Texture <arcade.Texture>` instance. The chart is
    automatically redrawn to this internal
    :class:`Texture <arcade.Texture>` every ``update_rate`` seconds.

    Args:
        width (int):
            The width of the chart texture in pixels
        height (int):
            The height of the chart texture in pixels
        graph_data (str):
            The pyglet event handler or statistic to track
        update_rate (float):
            How often the graph updates, in seconds
        background_color (RGBA255):
            The background color of the chart
        data_line_color (RGBA255):
            Color of the line tracking drawn
        axis_color (RGBA255):
            The color to draw the x & y axes in
        font_color (RGBA255):
            The color of the label font
        font_size (int):
            The size of the label font in points
        y_axis_num_lines (int):
            How many grid lines should be used to divide the y scale of the graph.
        view_y_scale_step (float):
            The graph's view area will be scaled to a multiple of this value to
            fit to the data currently displayed.
    """

    def __init__(
        self,
        width: int,
        height: int,
        graph_data: str = "FPS",
        update_rate: float = 0.1,
        background_color: RGBA255 = arcade.color.BLACK,
        data_line_color: RGBA255 = arcade.color.WHITE,
        axis_color: RGBA255 = arcade.color.DARK_YELLOW,
        grid_color: RGBA255 = arcade.color.DARK_YELLOW,
        font_color: RGBA255 = arcade.color.WHITE,
        font_size: int = 10,
        y_axis_num_lines: int = 4,
        view_y_scale_step: float = 20.0,
    ):
        unique_id = str(random.random())
        self.minimap_texture = arcade.Texture.create_empty(unique_id, (width, height))
        super().__init__(self.minimap_texture)
        self.proj = 0, self.width, 0, self.height

        # The data line is redrawn each update by a function that does
        # not cache vertices, so there is no need to make this attribute
        # a property that updates geometry when set.
        self.line_color = Color.from_iterable(data_line_color)
        """The color of the line tracking the data."""

        # Store visual style info for cached pyglet shape geometry
        self._background_color = Color.from_iterable(background_color)
        self._grid_color = Color.from_iterable(grid_color)
        self._axis_color = Color.from_iterable(axis_color)
        self._font_color = Color.from_iterable(font_color)
        self._font_size = font_size
        self._y_axis_num_lines = y_axis_num_lines
        self._left_x = 25
        self._bottom_y = 15

        # Variables for rendering the data line
        self.graph_data = graph_data
        """The graphed data type, either "FPS" or a pyglet event handler name."""

        self._data_to_graph: list[float] = []
        self._view_max_value = 0.0  # We'll calculate this once we have data
        self._view_y_scale_step = view_y_scale_step
        self._view_height = self._texture.height - self._bottom_y  # type: ignore
        self._y_increment = self._view_height / self._y_axis_num_lines

        # Set up internal Text object & line caches

        self._pyglet_batch = Batch()  # Used to draw graph elements

        # Convenient storage for iteration during color updates
        self._vertical_axis_text_objects: list[arcade.Text] = []
        self._all_text_objects: list[arcade.Text] = []
        self._grid_lines: list[Line] = []

        # Create the bottom label text object
        self._bottom_label = arcade.Text(
            graph_data,
            0,
            2,
            self._font_color,
            self._font_size,
            align="center",
            width=int(width),
        )
        self._all_text_objects.append(self._bottom_label)

        # Create the axes
        self._x_axis = Line(
            self._left_x,
            self._bottom_y,
            self._left_x,
            height,
            batch=self._pyglet_batch,
            color=self._axis_color,
        )

        self._y_axis = Line(
            self._left_x,
            self._bottom_y,
            width,
            self._bottom_y,
            batch=self._pyglet_batch,
            color=self._axis_color,
        )

        # Create the Y scale text objects & lines
        for i in range(self._y_axis_num_lines):
            y_level = self._bottom_y + self._y_increment * i
            self._vertical_axis_text_objects.append(
                arcade.Text(
                    "0",  # Ensure the lowest y axis label is always 0
                    self._left_x,
                    int(y_level),
                    self._font_color,
                    self._font_size,
                    anchor_x="right",
                    anchor_y="center",
                )
            )
            self._grid_lines.append(
                Line(
                    self._left_x,
                    y_level,
                    width,
                    y_level,
                    batch=self._pyglet_batch,
                    color=self._grid_color,
                )
            )

        self._all_text_objects.extend(self._vertical_axis_text_objects)

        # Enable auto-update
        pyglet.clock.schedule_interval(self.update_graph, update_rate)

    @property
    def background_color(self) -> Color:
        """Get or set the background color of the graph."""
        return self._background_color

    @background_color.setter
    def background_color(self, new_color: RGBA255):
        self._background_color = Color.from_iterable(new_color)

    @property
    def grid_color(self) -> Color:
        """Get or set the color of the grid lines."""
        return self._grid_color

    @grid_color.setter
    def grid_color(self, raw_color: RGBA255):
        new_color = Color.from_iterable(raw_color)
        for grid_line in self._grid_lines:
            grid_line.color = new_color

    @property
    def axis_color(self) -> Color:
        """Get or set the color of the x and y axes."""
        return self._axis_color

    @axis_color.setter
    def axis_color(self, raw_color: RGBA255):
        new_color = Color.from_iterable(raw_color)
        self._x_axis.color = new_color
        self._y_axis.color = new_color

    @property
    def font_size(self) -> int:
        """Get or set the font size of the labels."""
        return self._font_size

    @font_size.setter
    def font_size(self, new: int):
        self._font_size = new
        for text in self._all_text_objects:
            text.font_size = new

    @property
    def font_color(self) -> Color:
        """Get or set the font color of the labels."""
        return self._font_color

    @font_color.setter
    def font_color(self, raw_color: Color):
        new_color = Color.from_iterable(raw_color)
        self._font_color = new_color
        for text in self._all_text_objects:
            text.color = new_color

    def remove_from_sprite_lists(self) -> None:
        """
        Remove the sprite from all lists and cancel the update event.
        """
        super().remove_from_sprite_lists()

        # It is very important to call this to prevent potential
        # issues such as crashes or excess memory use from failed
        # garbage collection.
        pyglet.clock.unschedule(self.update)

    def update_graph(self, delta_time: float) -> None:
        """
        Update the graph by redrawing the internal texture data.

        .. warning:: You do not need to call this method! It will be
                     called automatically!

        Args:
            delta_time: Elapsed time in seconds. Passed by the pyglet
                        scheduler.
        """

        # Skip update if there is no SpriteList that can draw this graph
        if self.sprite_lists is None or len(self.sprite_lists) == 0:
            return

        sprite_list = self.sprite_lists[0]
        atlas = sprite_list.atlas

        # Clear and return if timings are disabled
        if not arcade.timings_enabled():
            # Please forgive the ugly spacing. It makes type checking work.
            with atlas.render_into(self.minimap_texture, projection=self.proj) as fbo:  # type: ignore
                fbo.clear(color=(0, 0, 0, 255))
            return

        # Get FPS and add to our historical data
        data_to_graph = self._data_to_graph
        graph_data = self.graph_data
        if graph_data == "FPS":
            data_to_graph.append(arcade.get_fps())
        else:
            timings = arcade.get_timings()
            if graph_data in timings:
                timing_list = timings[self.graph_data]
                avg_timing = sum(timing_list) / len(timing_list)
                data_to_graph.append(avg_timing * 1000)

        # Skip update if there is no data to graph
        if len(data_to_graph) == 0:
            return

        # Using locals for frequently used values is faster than
        # looking up instance variables repeatedly.
        bottom_y = self._bottom_y
        left_x = self._left_x
        view_y_scale_step = self._view_y_scale_step
        vertical_axis_text_objects = self._vertical_axis_text_objects
        view_height = self._view_height

        # We have to render at the internal texture's original size to
        # prevent distortion and bugs when the sprite is scaled.
        texture_width, texture_height = self._texture.size  # type: ignore

        # Toss old data by removing leftmost entries
        while len(data_to_graph) > texture_width - left_x:
            data_to_graph.pop(0)

        # Calculate the value at the top of the chart
        max_value = max(data_to_graph)
        view_max_value = ((max_value + 1.5) // view_y_scale_step + 1) * view_y_scale_step

        # Calculate draw positions of each pixel on the data line
        point_list = []
        x = left_x
        for reading in data_to_graph:
            y = (reading / view_max_value) * view_height + bottom_y
            point_list.append((x, y))
            x += 1

        # Update the view scale & labels if needed
        if view_max_value != self._view_max_value:
            self._view_max_value = view_max_value
            view_y_legend_increment = self._view_max_value // self._y_axis_num_lines
            for index in range(1, len(vertical_axis_text_objects)):
                text_object = vertical_axis_text_objects[index]
                text_object.text = f"{int(index * view_y_legend_increment)}"

        # Render to the internal texture
        # This ugly spacing is intentional to make type checking work.
        with atlas.render_into(self.minimap_texture, projection=self.proj) as fbo:  # type: ignore
            # Set the background color
            fbo.clear(color=self.background_color)

            # Draw lines & their labels
            for text in self._all_text_objects:
                text.draw()
            self._pyglet_batch.draw()

            # Draw the data line
            arcade.draw_line_strip(point_list, self.line_color)



from pathlib import Path

from PIL import Image

from arcade.hitbox import HitBoxAlgorithm
from arcade.resources import resolve

from .spritesheet import SpriteSheet
from .texture import ImageData, Texture


def load_texture(
    file_path: str | Path,
    *,
    hit_box_algorithm: HitBoxAlgorithm | None = None,
    hash: str | None = None,
) -> Texture:
    """
    Load a texture from disk (no caching).

    When loading a texture a set of hit box points will be generated by default
    based on the pixel data. The default hit box algorithm is a simple 4 to 8
    point hit box. This can be overridden by passing a different hit box or
    the global default can be set in the hitbox module.

    Examples::

        # We can load a texture using resource handles, string path or Path object
        texture = load_texture(":resources:image.png")
        texture = load_texture("image.png")
        texture = load_texture(Path("image.png"))

        # We can also specify a hit box algorithm to use for this texture
        texture = load_texture(
            ":resources:images/enemies/slimeBlock.png",
            hit_box_algorithm=arcade.hitbox.algo_detailed.
        )

    Args:
        file_path:
            Path to the image file
        hit_box_algorithm:
            The hit box algorithm to use for this texture. If not specified
            the global default will be used.
        hash:
            (advanced) Optional custom hash/name for the loaded image.
            This is used for texture caching and global uniqueness
            in texture atlases.
    """
    if isinstance(file_path, str):
        file_path = resolve(file_path)

    im: Image.Image = Image.open(file_path)  # type: ignore
    if im.mode != "RGBA":
        im = im.convert("RGBA")

    im_data = ImageData(im, hash=hash)
    tex = Texture(im_data, hit_box_algorithm=hit_box_algorithm)
    tex.file_path = file_path
    return tex


def load_image(
    file_path: str | Path,
    *,
    mode: str = "RGBA",
) -> Image.Image:
    """
    Load a Pillow image from disk (no caching).

    Normally you would use ``load_texture`` instead of this function.
    This function is useful when you want to load an image and then
    manipulate it before creating a texture.

    Note that Arcade mainly works with RGBA images. If you override
    the mode you might need to convert the final image to RGBA.

    Args:
        file_path:
            Path to the image file
        mode:
            The desired mode for the image (default: "RGBA")
    """
    if isinstance(file_path, str):
        file_path = resolve(file_path)

    im: Image.Image = Image.open(file_path)  # type: ignore
    if im.mode != mode:
        im = im.convert(mode)

    return im


def load_spritesheet(file_name: str | Path) -> SpriteSheet:
    """
    Loads an image from disk returning a sprite sheet that can
    further be used to slice out smaller images.

    Args:
        file_name: Path to the image file
    """
    if isinstance(file_name, str):
        file_name = resolve(file_name)

    im: Image.Image = Image.open(file_name)
    if im.mode != "RGBA":
        im = im.convert("RGBA")

    return SpriteSheet.from_image(im)



"""
Moving Sprite Stress Test

Simple program to test how fast we can draw sprites that are moving

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.stress_test_draw_moving
"""
import arcade
import random
import timeit
import time
import collections
import pyglet

# --- Constants ---
SPRITE_SCALING_COIN = 0.09
SPRITE_SCALING_PLAYER = 0.5
SPRITE_NATIVE_SIZE = 128
SPRITE_SIZE = int(SPRITE_NATIVE_SIZE * SPRITE_SCALING_COIN)
COIN_COUNT_INCREMENT = 500

STOP_COUNT = 12000

WINDOW_WIDTH = 1800
WINDOW_HEIGHT = 1000
WINDOW_TITLE = "Moving Sprite Stress Test"

USE_SPATIAL_HASHING = True
if USE_SPATIAL_HASHING:
    RESULTS_FILE = "stress_test_collision_arcade_spatial.csv"
else:
    RESULTS_FILE = "stress_test_collision_arcade.csv"


class FPSCounter:
    def __init__(self):
        self.time = time.perf_counter()
        self.frame_times = collections.deque(maxlen=60)

    def tick(self):
        t1 = time.perf_counter()
        dt = t1 - self.time
        self.time = t1
        self.frame_times.append(dt)

    def get_fps(self):
        total_time = sum(self.frame_times)
        if total_time == 0:
            return 0
        else:
            return len(self.frame_times) / sum(self.frame_times)


class GameView(arcade.View):
    """ Our custom Window Class"""

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.coin_list = None
        self.player_list = None
        self.player = None

        self.processing_time = 0
        self.draw_time = 0
        self.program_start_time = timeit.default_timer()
        self.sprite_count_list = []
        self.fps_list = []
        self.processing_time_list = []
        self.drawing_time_list = []
        self.last_fps_reading = 0
        self.fps = FPSCounter()

        self.background_color = arcade.color.AMAZON

        # Open file to save timings
        self.results_file = open(RESULTS_FILE, "w")

    def add_coins(self):

        # Create the coins
        for i in range(COIN_COUNT_INCREMENT):
            # Create the coin instance
            # Coin image from kenney.nl
            coin = arcade.Sprite(":resources:images/items/coinGold.png", SPRITE_SCALING_COIN)

            # Position the coin
            coin.center_x = random.randrange(SPRITE_SIZE, WINDOW_WIDTH - SPRITE_SIZE)
            coin.center_y = random.randrange(SPRITE_SIZE, WINDOW_HEIGHT - SPRITE_SIZE)

            # Add the coin to the lists
            self.coin_list.append(coin)

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.coin_list = arcade.SpriteList(use_spatial_hash=USE_SPATIAL_HASHING)
        self.player_list = arcade.SpriteList()
        self.player = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            SPRITE_SCALING_PLAYER,
        )
        self.player.center_x = random.randrange(WINDOW_WIDTH)
        self.player.center_y = random.randrange(WINDOW_HEIGHT)
        self.player.change_x = 3
        self.player.change_y = 5
        self.player_list.append(self.player)

    def on_draw(self):
        """ Draw everything """

        # Start timing how long this takes
        draw_start_time = timeit.default_timer()

        self.clear()
        self.coin_list.draw()
        self.player_list.draw()

        # Display info on sprites
        output = f"Sprite count: {len(self.coin_list):,}"
        arcade.draw_text(output, 20, WINDOW_HEIGHT - 20, arcade.color.BLACK, 16)

        # Display timings
        output = f"Processing time: {self.processing_time:.3f}"
        arcade.draw_text(output, 20, WINDOW_HEIGHT - 40, arcade.color.BLACK, 16)

        output = f"Drawing time: {self.draw_time:.3f}"
        arcade.draw_text(output, 20, WINDOW_HEIGHT - 60, arcade.color.BLACK, 16)

        fps = self.fps.get_fps()
        output = f"FPS: {fps:3.0f}"
        arcade.draw_text(output, 20, WINDOW_HEIGHT - 80, arcade.color.BLACK, 16)

        self.draw_time = timeit.default_timer() - draw_start_time
        self.fps.tick()

    def on_update(self, delta_time):
        # Start update timer

        start_time = timeit.default_timer()

        self.player_list.update()
        if self.player.center_x < 0 and self.player.change_x < 0:
            self.player.change_x *= -1
        if self.player.center_y < 0 and self.player.change_y < 0:
            self.player.change_y *= -1

        if self.player.center_x > WINDOW_WIDTH and self.player.change_x > 0:
            self.player.change_x *= -1
        if self.player.center_y > WINDOW_HEIGHT and self.player.change_y > 0:
            self.player.change_y *= -1

        coin_hit_list = arcade.check_for_collision_with_list(self.player, self.coin_list)
        for coin in coin_hit_list:
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(WINDOW_HEIGHT)

        # Save the time it took to do this.
        self.processing_time = timeit.default_timer() - start_time

        # Total time program has been running
        total_program_time = int(timeit.default_timer() - self.program_start_time)

        # Print out stats, or add more sprites
        if total_program_time > self.last_fps_reading:
            self.last_fps_reading = total_program_time

            # It takes the program a while to "warm up", so the first
            # few seconds our readings will be off. So wait some time
            # before taking readings
            if total_program_time > 5:

                # We want the program to run for a while before taking
                # timing measurements. We don't want the time it takes
                # to add new sprites to be part of that measurement. So
                # make sure we have a clear second of nothing but
                # running the sprites, and not adding the sprites.
                if total_program_time % 2 == 1:

                    output = (
                        f"{total_program_time}, {len(self.coin_list)}, {self.fps.get_fps():.1f}, "
                        f"{self.processing_time:.4f}, {self.draw_time:.4f}\n"
                    )
                    print(output, end="")
                    self.results_file.write(output)

                    if len(self.coin_list) >= STOP_COUNT:
                        self.results_file.close()
                        pyglet.app.exit()
                        return

                    # Take timings
                    print(
                        f"{total_program_time}, {len(self.coin_list)}, {self.fps.get_fps():.1f}, "
                        f"{self.processing_time:.4f}, {self.draw_time:.4f}"
                    )
                    self.sprite_count_list.append(len(self.coin_list))
                    self.fps_list.append(round(self.fps.get_fps(), 1))
                    self.processing_time_list.append(self.processing_time)
                    self.drawing_time_list.append(self.draw_time)

                    # Now add the coins
                    self.add_coins()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



import sys
import traceback
from collections.abc import Callable
from typing import Any, Generic, TypeVar, cast
from weakref import WeakKeyDictionary, ref

from typing_extensions import Self, overload, override

P = TypeVar("P")


class _Obs(Generic[P]):
    """
    Internal holder for Property value and change listeners
    """

    __slots__ = ("value", "listeners")

    def __init__(self, value: P):
        self.value = value
        # This will keep any added listener even if it is not referenced anymore
        # and would be garbage collected
        self.listeners: set[Callable[[Any, P], Any] | Callable[[], Any]] = set()


class Property(Generic[P]):
    """An observable property which triggers observers when changed.

    .. code-block:: python

        def log_change(instance, value):
            print("Something changed")

        class MyObject:
            name = Property()

        my_obj = MyObject()
        bind(my_obj, "name", log_change)
        unbind(my_obj, "name", log_change)

        my_obj.name = "Hans"
        # > Something changed

    Properties provide a less verbose way to implement the observer pattern in comparison to
    using the `property` decorator.

    Args:
        default: Default value which is returned, if no value set before
        default_factory: A callable which returns the default value.
            Will be called with the property and the instance
    """

    __slots__ = ("name", "default_factory", "obs")
    name: str
    """Attribute name of the property"""
    default_factory: Callable[[Any, Any], P]
    """Default factory to create the initial value"""
    obs: WeakKeyDictionary[Any, _Obs]
    """Weak dictionary to hold the value and listeners"""

    def __init__(
        self,
        default: P | None = None,
        default_factory: Callable[[Any, Any], P] | None = None,
    ):
        if default_factory is None:
            default_factory = lambda prop, instance: cast(P, default)

        self.default_factory = default_factory
        self.obs: WeakKeyDictionary[Any, _Obs] = WeakKeyDictionary()

    def _get_obs(self, instance) -> _Obs:
        obs = self.obs.get(instance)
        if obs is None:
            obs = _Obs(self.default_factory(self, instance))
            self.obs[instance] = obs
        return obs

    def get(self, instance) -> P:
        """Get value for owner instance"""
        obs = self._get_obs(instance)
        return obs.value

    def set(self, instance, value):
        """Set value for owner instance"""
        obs = self._get_obs(instance)
        if obs.value != value:
            obs.value = value
            self.dispatch(instance, value)

    def dispatch(self, instance, value):
        """Notifies every listener, which subscribed to the change event.

        Args:
            instance: Property instance
            value: new value to set

        """
        obs = self._get_obs(instance)
        for listener in obs.listeners:
            try:
                try:
                    # FIXME if listener() raises an error, the invalid call will be
                    #       also shown as an exception
                    listener(instance, value)  # type: ignore
                except TypeError:
                    # If the listener does not accept arguments, we call it without it
                    listener()  # type: ignore
            except Exception:
                print(
                    f"Change listener for {instance}.{self.name} = {value} raised an exception!",
                    file=sys.stderr,
                )
                traceback.print_exc()

    def bind(self, instance, callback):
        """Binds a function to the change event of the property.

        A reference to the function will be kept.

        Args:
             instance: The instance to bind the callback to.
             callback: The callback to bind.
        """
        obs = self._get_obs(instance)
        # Instance methods are bound methods, which can not be referenced by normal `ref()`
        # if listeners would be a WeakSet, we would have to add listeners as WeakMethod
        # ourselves into `WeakSet.data`.
        obs.listeners.add(callback)

    def unbind(self, instance, callback):
        """Unbinds a function from the change event of the property.

        Args:
            instance: The target instance.
            callback: The callback to unbind.
        """
        obs = self._get_obs(instance)
        obs.listeners.remove(callback)

    def __set_name__(self, owner, name):
        self.name = name

    @overload
    def __get__(self, instance: None, instance_type) -> Self: ...

    @overload
    def __get__(self, instance: Any, instance_type) -> P: ...

    def __get__(self, instance: Any | None, instance_type) -> Self | P:
        if instance is None:
            return self
        return self.get(instance)

    def __set__(self, instance, value: P):
        self.set(instance, value)


def bind(instance, property: str, callback):
    """Bind a function to the change event of the property.

    A reference to the function will be kept, so that it will be still
    invoked even if it would normally have been garbage collected:

    .. code-block:: python

        def log_change(instance, value):
            print(f"Value of {instance} changed to {value}")

        class MyObject:
            name = Property()

        my_obj = MyObject()
        bind(my_obj, "name", log_change)

        my_obj.name = "Hans"
        # > Value of <__main__.MyObject ...> changed to Hans

    Args:
        instance: Instance owning the property
        property: Name of the property
        callback: Function to call

    Returns:
        None
    """
    t = type(instance)
    prop = getattr(t, property)
    if isinstance(prop, Property):
        prop.bind(instance, callback)


def unbind(instance, property: str, callback):
    """Unbinds a function from the change event of the property.

    .. code-block:: python

        def log_change(instance, value):
            print("Something changed")

        class MyObject:
            name = Property()

        my_obj = MyObject()
        bind(my_obj, "name", log_change)
        unbind(my_obj, "name", log_change)

        my_obj.name = "Hans"
        # > Something changed

    Args:
        instance: Instance owning the property
        property: Name of the property
        callback: Function to unbind

    Returns:
        None
    """
    t = type(instance)
    prop = getattr(t, property)
    if isinstance(prop, Property):
        prop.unbind(instance, callback)


class _ObservableDict(dict):
    """Internal class to observe changes inside a native python dict."""

    __slots__ = ("prop", "obj")

    def __init__(self, prop: Property, instance, *args):
        self.prop: Property = prop
        self.obj = ref(instance)
        super().__init__(*args)

    def dispatch(self):
        self.prop.dispatch(self.obj(), self)

    @override
    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
        self.dispatch()

    @override
    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self.dispatch()

    @override
    def clear(self):
        dict.clear(self)
        self.dispatch()

    @override
    def pop(self, *args):
        result = dict.pop(self, *args)
        self.dispatch()
        return result

    @override
    def popitem(self):
        result = dict.popitem(self)
        self.dispatch()
        return result

    @override
    def setdefault(self, *args):
        dict.setdefault(self, *args)
        self.dispatch()

    @override
    def update(self, *args):
        dict.update(self, *args)
        self.dispatch()


K = TypeVar("K")
V = TypeVar("V")


class DictProperty(Property[dict[K, V]], Generic[K, V]):
    """Property that represents a dict.

    Only dict are allowed. Any other classes are forbidden.
    """

    def __init__(self):
        super().__init__(default_factory=_ObservableDict)

    @override
    def set(self, instance, value: dict):
        """Set value for owner instance, wraps the dict into an observable dict."""
        value = _ObservableDict(self, instance, value)
        super().set(instance, value)


class _ObservableList(list):
    """Internal class to observe changes inside a native python list.

    Args:
        prop: Property instance
        instance: Instance owning the property
        *args: List of arguments to pass to the list
    """

    __slots__ = ("prop", "obj")

    def __init__(self, prop: Property, instance, *args):
        self.prop: Property = prop
        self.obj = ref(instance)
        super().__init__(*args)

    def dispatch(self):
        """Dispatches the change event."""
        self.prop.dispatch(self.obj(), self)

    @override
    def __setitem__(self, key, value):
        list.__setitem__(self, key, value)
        self.dispatch()

    @override
    def __delitem__(self, key):
        list.__delitem__(self, key)
        self.dispatch()

    @override
    def __iadd__(self, *args):
        list.__iadd__(self, *args)
        self.dispatch()
        return self

    @override
    def __imul__(self, *args):
        list.__imul__(self, *args)
        self.dispatch()
        return self

    @override
    def append(self, *args):
        """Proxy for list.append() which dispatches the change event."""
        list.append(self, *args)
        self.dispatch()

    @override
    def clear(self):
        """Proxy for list.clear() which dispatches the change event."""
        list.clear(self)
        self.dispatch()

    @override
    def remove(self, *args):
        """Proxy for list.remove() which dispatches the change event."""
        list.remove(self, *args)
        self.dispatch()

    @override
    def insert(self, *args):
        """Proxy for list.insert() which dispatches the change event."""
        list.insert(self, *args)
        self.dispatch()

    @override
    def pop(self, *args):
        """Proxy for list.pop() which dispatches the change"""
        result = list.pop(self, *args)
        self.dispatch()
        return result

    @override
    def extend(self, *args):
        """Proxy for list.extend() which dispatches the change event."""
        list.extend(self, *args)
        self.dispatch()

    @override
    def sort(self, **kwargs):
        """Proxy for list.sort() which dispatches the change event."""
        list.sort(self, **kwargs)
        self.dispatch()

    @override
    def reverse(self):
        """Proxy for list.reverse() which dispatches the change event."""
        list.reverse(self)
        self.dispatch()


class ListProperty(Property[list[P]], Generic[P]):
    """Property that represents a list.

    Only list are allowed. Any other classes are forbidden.
    """

    def __init__(self):
        super().__init__(default_factory=_ObservableList)

    @override
    def set(self, instance, value: list):
        """Set value for owner instance, wraps the list into an observable list."""
        value = _ObservableList(self, instance, value)
        super().set(instance, value)



import pymunk
from PIL.Image import Image
from pymunk import Vec2d
from pymunk.autogeometry import (
    PolylineSet,
    march_soft,
    simplify_curves,
)

from arcade.types import RGBA255, Point2, Point2List

from .base import HitBoxAlgorithm


class PymunkHitBoxAlgorithm(HitBoxAlgorithm):
    """
    Hit box point algorithm that uses pymunk to calculate the points.

    This is a more accurate algorithm generating more points. The
    point count can be controlled with the ``detail`` parameter.
    """

    #: The default detail when creating a new instance.
    default_detail = 4.5

    def __init__(self, *, detail: float | None = None):
        super().__init__()
        self.detail = detail or self.default_detail
        self._cache_name += f"|detail={self.detail}"

    def __call__(self, *, detail: float | None = None) -> "PymunkHitBoxAlgorithm":
        """Create a new instance with new default values"""
        return PymunkHitBoxAlgorithm(detail=detail or self.detail)

    def calculate(self, image: Image, detail: float | None = None, **kwargs) -> Point2List:
        """
        Given an RGBA image, this returns points that make up a hit box around it.

        Args:
            image: Image get hit box from.
            detail: How detailed to make the hit box. There's a
                trade-off in number of points vs. accuracy.
        """
        hit_box_detail = detail or self.detail

        if image.mode != "RGBA":
            raise ValueError("Image mode is not RGBA. image.convert('RGBA') is needed.")

        # Trace the image finding all the outlines and holes
        line_sets = self.trace_image(image)
        if len(line_sets) == 0:
            return self.create_bounding_box(image)

        # Get the largest line set
        line_set = self.select_largest_line_set(line_sets)

        # Reduce number of vertices
        if len(line_set) > 4:
            line_set = simplify_curves(line_set, hit_box_detail)

        return self.to_points_list(image, line_set)

    def to_points_list(self, image: Image, line_set: list[Vec2d]) -> Point2List:
        """
        Convert a line set to a list of points.

        Coordinates are offset so ``(0,0)`` is the center of the image.

        Args:
            image: Image to trace.
            line_set: Line set to convert.
        """
        # Convert to normal points, offset fo 0,0 is center, flip the y
        hh = image.height / 2.0
        hw = image.width / 2.0
        points = []
        for vec2 in line_set:
            point_tuple = (
                float(round(vec2.x - hw)),
                float(round(image.height - (vec2.y - hh) - image.height)),
            )
            points.append(point_tuple)

        # Remove duplicate end point
        if len(points) > 1 and points[0] == points[-1]:
            points.pop()

        # Return immutable data
        return tuple(points)

    def trace_image(self, image: Image) -> PolylineSet:
        """
        Trace the image and return a :py:class:~collections.abc.Sequence` of line sets.

        .. important:: The image :py:attr:`~PIL.Image.Image.mode` must be ``"RGBA"``!

                       * This method raises a :py:class:`TypeError` when it isn't
                       * Use :py:meth:`convert("RGBA") <PIL.Image.Image.convert>` to
                         convert

        The returned object will be a :py:mod:`pymunk`
        :py:class:`~pymunk.autogeometry.PolylineSet`. Each
        :py:class:`list` inside it will contain points as
        :py:class:`pymunk.vec2d.Vec2d` instances. These lists
        may represent:

        * the outline of the image's contents
        * the holes in the image

        When this method returns more than one line set,
        it's important to pick the one which covers the largest
        portion of the image.

        Args:
            image: A :py:class:`PIL.Image.Image` to trace.

        Returns:
            A :py:mod:`pymunk` object which is a :py:class:`~collections.abc.Sequence`
            of :py:class:`~pymunk.autogeometry.PolylineSet` of line sets.
        """
        if image.mode != "RGBA":
            raise ValueError("Image's mode!='RGBA'! Try using image.convert(\"RGBA\").")

        def sample_func(sample_point: Point2) -> int:
            """Function used to sample image."""
            # Return 0 when outside of bounds
            if (
                sample_point[0] < 0
                or sample_point[1] < 0
                or sample_point[0] >= image.width
                or sample_point[1] >= image.height
            ):
                return 0

            point_tuple = int(sample_point[0]), int(sample_point[1])
            color: RGBA255 = image.getpixel(point_tuple)  # type: ignore

            return 255 if color[3] > 0 else 0

        # Do a quick check if it is a full tile
        # Points are pixel coordinates
        px1 = 0, 0
        px2 = 0, image.height - 1
        px3 = image.width - 1, image.height - 1
        px4 = image.width - 1, 0
        if sample_func(px1) and sample_func(px2) and sample_func(px3) and sample_func(px4):
            # Actual points are in world coordinates
            p1 = 0.0, 0.0
            p2 = 0.0, float(image.height)
            p3 = float(image.width), float(image.height)
            p4 = float(image.width), 0.0
            # Manually create a line set
            points = PolylineSet()
            points.collect_segment(p2, p3)
            points.collect_segment(p3, p4)
            points.collect_segment(p4, p1)
            return points

        # Get the bounding box
        logo_bb = pymunk.BB(-1, -1, image.width, image.height)

        # How often to sample?
        down_res = 1
        horizontal_samples = int(image.width / down_res)
        vertical_samples = int(image.height / down_res)

        # Run the trace
        # Get back one or more sets of lines covering stuff.
        # We want the one that covers the most of the sprite
        # or the line set might just be a hole in the sprite.
        return march_soft(logo_bb, horizontal_samples, vertical_samples, 99, sample_func)

    def select_largest_line_set(self, line_sets: PolylineSet) -> list[Vec2d]:
        """
        Given a list of line sets, return the one that covers the most of the image.

        Args:
            line_sets: List of line sets.
        """
        if len(line_sets) == 1:
            return line_sets[0]

        # We have more than one line set.
        # Try and find one that covers most of the sprite.
        selected_line_set = line_sets[0]
        largest_area = -1.0

        for line_set in line_sets:
            min_x = None
            min_y = None
            max_x = None
            max_y = None
            for point in line_set:
                if min_x is None or point.x < min_x:
                    min_x = point.x
                if max_x is None or point.x > max_x:
                    max_x = point.x
                if min_y is None or point.y < min_y:
                    min_y = point.y
                if max_y is None or point.y > max_y:
                    max_y = point.y

            if min_x is None or max_x is None or min_y is None or max_y is None:
                raise ValueError("No points in bounding box.")

            # Calculate the area of the bounding box
            area = (max_x - min_x) * (max_y - min_y)
            if area > largest_area:
                largest_area = area
                selected_line_set = line_set

        return selected_line_set
